#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Kernel decompositon into features and coefficients based on Gauss-Christoffel
quadrature aka Gaussian quadrature.
"""

import warnings
from typing import Mapping, Tuple, Type

import tensorflow as tf

import gpflow
from gpflow.base import TensorType
from gpflow.quadrature.gauss_hermite import ndgh_points_and_weights

from gpflux.layers.basis_functions.fourier_features.base import FourierFeaturesBase
from gpflux.layers.basis_functions.fourier_features.utils import _bases_concat
from gpflux.types import ShapeType

"""
Kernels supported by :class:`QuadratureFourierFeatures`.

Currently we only support the :class:`gpflow.kernels.SquaredExponential` kernel.
For Matern kernels please use :class:`RandomFourierFeatures`
or :class:`RandomFourierFeaturesCosine`.
"""
QFF_SUPPORTED_KERNELS: Tuple[Type[gpflow.kernels.Stationary], ...] = (
    gpflow.kernels.SquaredExponential,
)


class QuadratureFourierFeatures(FourierFeaturesBase):
    def __init__(self, kernel: gpflow.kernels.Kernel, n_components: int, **kwargs: Mapping):
        assert isinstance(kernel, QFF_SUPPORTED_KERNELS), "Unsupported Kernel"
        if tf.reduce_any(tf.less(kernel.lengthscales, 1e-1)):
            warnings.warn(
                "Quadrature Fourier feature approximation of kernels "
                "with small lengthscale lead to unexpected behaviors!"
            )
        super(QuadratureFourierFeatures, self).__init__(kernel, n_components, **kwargs)

    def build(self, input_shape: ShapeType) -> None:
        """
        Creates the variables of the layer.
        See `tf.keras.layers.Layer.build()
        <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#build>`_.
        """
        input_dim = input_shape[-1]
        abscissa_value, omegas_value = ndgh_points_and_weights(
            dim=input_dim, n_gh=self.n_components
        )
        omegas_value = tf.squeeze(omegas_value, axis=-1)

        # Quadrature node points
        self.abscissa = tf.Variable(initial_value=abscissa_value, trainable=False)  # (M^D, D)
        # Gauss-Hermite weights
        self.factors = tf.Variable(initial_value=omegas_value, trainable=False)  # (M^D,)
        super(QuadratureFourierFeatures, self).build(input_shape)

    def _compute_output_dim(self, input_shape: ShapeType) -> int:
        input_dim = input_shape[-1]
        return 2 * self.n_components ** input_dim

    def _compute_bases(self, inputs: TensorType) -> tf.Tensor:
        """
        Compute basis functions.

        :return: A tensor with the shape ``[N, 2M^D]``.
        """
        return _bases_concat(inputs, self.abscissa)

    def _compute_constant(self) -> tf.Tensor:
        """
        Compute normalizing constant for basis functions.

        :return: A tensor with the shape ``[2M^D,]``
        """
        return tf.tile(tf.sqrt(self.kernel.variance * self.factors), multiples=[2])



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Support for the `gpflow.optimizers.NaturalGradient` optimizer within Keras models.
"""

from typing import Any, List, Mapping, Optional, Tuple, Union

import tensorflow as tf
from tensorflow.python.util.object_identity import ObjectIdentitySet

import gpflow
from gpflow import Parameter
from gpflow.models.model import MeanAndVariance
from gpflow.optimizers import NaturalGradient

from gpflux.layers.gp_layer import GPLayer

__all__ = [
    "NatGradModel",
    "NatGradWrapper",
]


class NatGradModel(tf.keras.Model):
    r"""
    This is a drop-in replacement for `tf.keras.Model` when constructing GPflux
    models using the functional Keras style, to make it work with the
    NaturalGradient optimizers for q(u) distributions in GP layers.

    You must set the `natgrad_layers` property before compiling the model. Set it
    to the list of all :class:`~gpflux.layers.GPLayer`\ s you want to train using
    natural gradients. You can also set it to `True` to include all of them.

    This model's :meth:`compile` method has to be passed a list of optimizers, which
    must be one `gpflow.optimizers.NaturalGradient` instance per natgrad-trained
    :class:`~gpflux.layers.GPLayer`, followed by a regular optimizer (e.g.
    `tf.keras.optimizers.Adam`) as the last element to handle all other
    parameters (hyperparameters, inducing point locations).
    """

    @property
    def natgrad_layers(self) -> List[GPLayer]:
        """
        The list of layers in this model that should be optimized using
        `~gpflow.optimizers.NaturalGradient`.

        :getter: Returns a list of the layers that should be trained using
            `~gpflow.optimizers.NaturalGradient`.
        :setter: Sets the layers that should be trained using
            `~gpflow.optimizers.NaturalGradient`. Can be an explicit list or a `bool`:
            If set to `True`, it will select all `GPLayer` instances in the model layers.
        """
        if not hasattr(self, "_natgrad_layers"):
            raise AttributeError(
                f"natgrad_layers must be set before training {self.__class__.__name__}"
            )  # pragma: no cover
        return self._natgrad_layers

    @natgrad_layers.setter
    def natgrad_layers(self, layers: Union[List[GPLayer], bool]) -> None:
        if isinstance(layers, bool):
            if layers:  # all (GP) layers
                self._natgrad_layers = [
                    layer for layer in self.layers if isinstance(layer, GPLayer)
                ]
            else:  # no layers
                self._natgrad_layers = []
        else:
            self._natgrad_layers = layers

    @property
    def natgrad_optimizers(self) -> List[gpflow.optimizers.NaturalGradient]:
        if not hasattr(self, "_all_optimizers"):
            raise AttributeError(
                "natgrad_optimizers accessed before optimizer being set"
            )  # pragma: no cover
        if self._all_optimizers is None:
            return None  # type: ignore
        return self._all_optimizers[:-1]

    @property
    def optimizer(self) -> tf.optimizers.Optimizer:
        """
        HACK to cope with Keras's callbacks such as
        :class:`~tf.keras.callbacks.ReduceLROnPlateau`
        and
        :class:`~tf.keras.callbacks.LearningRateScheduler`
        having been hardcoded for a single optimizer.
        """
        if not hasattr(self, "_all_optimizers"):
            raise AttributeError("optimizer accessed before being set")
        if self._all_optimizers is None:
            return None  # type: ignore
        return self._all_optimizers[-1]

    @optimizer.setter
    def optimizer(self, optimizers: List[Union[NaturalGradient, tf.optimizers.Optimizer]]) -> None:
        if optimizers is None:
            # tf.keras.Model.__init__() sets self.optimizer = None
            self._all_optimizers = None
            return

        if optimizers is self.optimizer:
            # Keras re-sets optimizer with itself; this should not have any effect on the state
            return

        if not isinstance(optimizers, (tuple, list)):
            raise TypeError(
                "`optimizer` needs to be a list of NaturalGradient optimizers for "
                "each element of `natgrad_layers` followed by one optimizer for all "
                "other parameters, "
                f"but was {optimizers}"
            )  # pragma: no cover
        if isinstance(optimizers[-1], NaturalGradient):
            raise TypeError(
                "The last element of the optimizer list applies to the non-variational "
                "parameters and cannot be a NaturalGradient optimizer, "
                f"but was {optimizers[-1]}"
            )  # pragma: no cover
        if not all(isinstance(o, NaturalGradient) for o in optimizers[:-1]):
            raise TypeError(
                "The all-but-last elements of the optimizer list must be "
                "NaturalGradient instances, one for each element of natgrad_layers, "
                f"but were {optimizers[:-1]}"
            )  # pragma: no cover
        self._all_optimizers = optimizers

    def _split_natgrad_params_and_other_vars(
        self,
    ) -> Tuple[List[Tuple[Parameter, Parameter]], List[tf.Variable]]:
        # NOTE the structure of variational_params is directly linked to the _natgrad_step,
        # do not change out of sync
        variational_params = [(layer.q_mu, layer.q_sqrt) for layer in self.natgrad_layers]
        # NOTE could use a natgrad_parameters attribute on a layer or a
        # singledispatch function to make this more flexible for other layers

        # Collect all trainable variables that are not part of variational_params:
        variational_vars_set = ObjectIdentitySet(
            p.unconstrained_variable for vp in variational_params for p in vp
        )
        other_vars = [v for v in self.trainable_variables if v not in variational_vars_set]

        return variational_params, other_vars

    def _apply_backwards_pass(self, loss: tf.Tensor, tape: tf.GradientTape) -> None:
        print("Executing NatGradModel backwards pass")
        # TODO(Ti) This is to check tf.function() compilation works and this
        # won't be called repeatedly. Surprisingly, it gets called *twice*...
        # Leaving the print here until we've established why twice, or whether
        # it's irrelevant, and that it all works correctly in practice.

        num_natgrad_layers = len(self.natgrad_layers)
        num_natgrad_opt = len(self.natgrad_optimizers)
        if num_natgrad_opt != num_natgrad_layers:
            raise ValueError(
                f"Model has {num_natgrad_opt} NaturalGradient optimizers, "
                f"but {num_natgrad_layers} variational distributions in "
                "natgrad_layers"
            )  # pragma: no cover

        variational_params, other_vars = self._split_natgrad_params_and_other_vars()
        variational_params_vars = [
            (q_mu.unconstrained_variable, q_sqrt.unconstrained_variable)
            for (q_mu, q_sqrt) in variational_params
        ]

        variational_params_grads, other_grads = tape.gradient(
            loss, (variational_params_vars, other_vars)
        )

        for (natgrad_optimizer, (q_mu_grad, q_sqrt_grad), (q_mu, q_sqrt)) in zip(
            self.natgrad_optimizers, variational_params_grads, variational_params
        ):
            natgrad_optimizer._natgrad_apply_gradients(q_mu_grad, q_sqrt_grad, q_mu, q_sqrt)

        self.optimizer.apply_gradients(zip(other_grads, other_vars))

    def train_step(self, data: Any) -> Mapping[str, Any]:
        """
        The logic for one training step. For more details of the
        implementation, see TensorFlow's documentation of how to
        `customize what happens in Model.fit
        <https://www.tensorflow.org/guide/keras/customizing_what_happens_in_fit>`_.
        """
        from tensorflow.python.keras.engine import data_adapter

        data = data_adapter.expand_1d(data)
        x, y, sample_weight = data_adapter.unpack_x_y_sample_weight(data)

        with tf.GradientTape() as tape:
            y_pred = self.__call__(x, training=True)
            loss = self.compiled_loss(y, y_pred, sample_weight, regularization_losses=self.losses)

        self._apply_backwards_pass(loss, tape=tape)

        self.compiled_metrics.update_state(y, y_pred, sample_weight)
        return {m.name: m.result() for m in self.metrics}


class NatGradWrapper(NatGradModel):
    """
    Wraps a class-based Keras model (e.g. the return value of
    `gpflux.models.DeepGP.as_training_model`) to make it work with
    `gpflow.optimizers.NaturalGradient` optimizers. For more details, see
    `NatGradModel`.

    (Note that you can also directly pass `NatGradModel` to the
    :class:`~gpflux.models.DeepGP`'s
    :attr:`~gpflux.models.DeepGP.default_model_class` or
    :meth:`~gpflux.models.DeepGP.as_training_model`'s *model_class* arguments.)

    .. todo::

        This class will probably be removed in the future.
    """

    def __init__(self, base_model: tf.keras.Model, *args: Any, **kwargs: Any):
        """
        :param base_model: the class-based Keras model to be wrapped
        """
        super().__init__(*args, **kwargs)
        self.base_model = base_model

    @property
    def layers(self) -> List[tf.keras.layers.Layer]:
        if not hasattr(self, "base_model"):
            # required for super().__init__(), in which base_model has not been set yet
            return super().layers
        else:
            return self.base_model.layers

    def call(self, data: Any, training: Optional[bool] = None) -> Union[tf.Tensor, MeanAndVariance]:
        """
        Calls the model on new inputs. Simply passes through to the ``base_model``.
        """
        return self.base_model.call(data, training=training)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides :class:`GPLayer`, which implements a Sparse Variational
Multioutput Gaussian Process as a Keras :class:`~tf.keras.layers.Layer`.
"""

import warnings
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

from gpflow import Parameter, default_float
from gpflow.base import TensorType
from gpflow.conditionals import conditional
from gpflow.inducing_variables import MultioutputInducingVariables
from gpflow.kernels import MultioutputKernel
from gpflow.kullback_leiblers import prior_kl
from gpflow.mean_functions import Identity, MeanFunction
from gpflow.utilities.bijectors import triangular

from gpflux.exceptions import GPLayerIncompatibilityException
from gpflux.math import _cholesky_with_jitter
from gpflux.runtime_checks import verify_compatibility
from gpflux.sampling.sample import Sample, efficient_sample


class GPLayer(tfp.layers.DistributionLambda):
    """
    A sparse variational multioutput GP layer. This layer holds the kernel,
    inducing variables and variational distribution, and mean function.
    """

    num_data: int
    """
    The number of points in the training dataset. This information is used to
    obtain the correct scaling between the data-fit and the KL term in the
    evidence lower bound (ELBO).
    """

    whiten: bool
    """
    This parameter determines the parameterisation of the inducing variables.

    If `True`, this layer uses the whitened (or non-centred) representation, in
    which (at the example of inducing point inducing variables) ``u = f(Z) =
    cholesky(Kuu) v``, and we parameterise an approximate posterior on ``v`` as
    ``q(v) = N(q_mu, q_sqrt q_sqrtᵀ)``. The prior on ``v`` is ``p(v) = N(0, I)``.

    If `False`, this layer uses the non-whitened (or centred) representation,
    in which we directly parameterise ``q(u) = N(q_mu, q_sqrt q_sqrtᵀ)``. The
    prior on ``u`` is ``p(u) = N(0, Kuu)``.
    """

    num_samples: Optional[int]
    """
    The number of samples drawn when coercing the output distribution of
    this layer to a `tf.Tensor`. (See :meth:`_convert_to_tensor_fn`.)
    """

    full_cov: bool
    """
    This parameter determines the behaviour of calling this layer. If `False`, only
    predict or sample marginals (diagonal of covariance) with respect to inputs.
    If `True`, predict or sample with the full covariance over the inputs.
    """

    full_output_cov: bool
    """
    This parameter determines the behaviour of calling this layer. If `False`, only
    predict or sample marginals (diagonal of covariance) with respect to outputs.
    If `True`, predict or sample with the full covariance over the outputs.
    """

    q_mu: Parameter
    r"""
    The mean of ``q(v)`` or ``q(u)`` (depending on whether :attr:`whiten`\ ed
    parametrisation is used).
    """

    q_sqrt: Parameter
    r"""
    The lower-triangular Cholesky factor of the covariance of ``q(v)`` or ``q(u)``
    (depending on whether :attr:`whiten`\ ed parametrisation is used).
    """

    def __init__(
        self,
        kernel: MultioutputKernel,
        inducing_variable: MultioutputInducingVariables,
        num_data: int,
        mean_function: Optional[MeanFunction] = None,
        *,
        num_samples: Optional[int] = None,
        full_cov: bool = False,
        full_output_cov: bool = False,
        num_latent_gps: int = None,
        whiten: bool = True,
        name: Optional[str] = None,
        verbose: bool = True,
    ):
        """
        :param kernel: The multioutput kernel for this layer.
        :param inducing_variable: The inducing features for this layer.
        :param num_data: The number of points in the training dataset (see :attr:`num_data`).
        :param mean_function: The mean function that will be applied to the
            inputs. Default: :class:`~gpflow.mean_functions.Identity`.

            .. note:: The Identity mean function requires the input and output
                dimensionality of this layer to be the same. If you want to
                change the dimensionality in a layer, you may want to provide a
                :class:`~gpflow.mean_functions.Linear` mean function instead.

        :param num_samples: The number of samples to draw when converting the
            :class:`~tfp.layers.DistributionLambda` into a `tf.Tensor`, see
            :meth:`_convert_to_tensor_fn`. Will be stored in the
            :attr:`num_samples` attribute.  If `None` (the default), draw a
            single sample without prefixing the sample shape (see
            :class:`tfp.distributions.Distribution`'s `sample()
            <https://www.tensorflow.org/probability/api_docs/python/tfp/distributions/Distribution#sample>`_
            method).
        :param full_cov: Sets default behaviour of calling this layer
            (:attr:`full_cov` attribute):
            If `False` (the default), only predict marginals (diagonal
            of covariance) with respect to inputs.
            If `True`, predict full covariance over inputs.
        :param full_output_cov: Sets default behaviour of calling this layer
            (:attr:`full_output_cov` attribute):
            If `False` (the default), only predict marginals (diagonal
            of covariance) with respect to outputs.
            If `True`, predict full covariance over outputs.
        :param num_latent_gps: The number of (latent) GPs in the layer
            (which can be different from the number of outputs, e.g. with a
            :class:`~gpflow.kernels.LinearCoregionalization` kernel).
            This is used to determine the size of the
            variational parameters :attr:`q_mu` and :attr:`q_sqrt`.
            If possible, it is inferred from the *kernel* and *inducing_variable*.
        :param whiten: If `True` (the default), uses the whitened parameterisation
            of the inducing variables; see :attr:`whiten`.
        :param name: The name of this layer.
        :param verbose: The verbosity mode. Set this parameter to `True`
            to show debug information.
        """

        super().__init__(
            make_distribution_fn=self._make_distribution_fn,
            convert_to_tensor_fn=self._convert_to_tensor_fn,
            dtype=default_float(),
            name=name,
        )

        self.kernel = kernel
        self.inducing_variable = inducing_variable

        self.num_data = num_data

        if mean_function is None:
            mean_function = Identity()
            if verbose:
                warnings.warn(
                    "Beware, no mean function was specified in the construction of the `GPLayer` "
                    "so the default `gpflow.mean_functions.Identity` is being used. "
                    "This mean function will only work if the input dimensionality "
                    "matches the number of latent Gaussian processes in the layer."
                )
        self.mean_function = mean_function

        self.full_output_cov = full_output_cov
        self.full_cov = full_cov
        self.whiten = whiten
        self.verbose = verbose

        try:
            num_inducing, self.num_latent_gps = verify_compatibility(
                kernel, mean_function, inducing_variable
            )
            # TODO: if num_latent_gps is not None, verify it is equal to self.num_latent_gps
        except GPLayerIncompatibilityException as e:
            if num_latent_gps is None:
                raise e

            if verbose:
                warnings.warn(
                    "Could not verify the compatibility of the `kernel`, `inducing_variable` "
                    "and `mean_function`. We advise using `gpflux.helpers.construct_*` to create "
                    "compatible kernels and inducing variables. As "
                    f"`num_latent_gps={num_latent_gps}` has been specified explicitly, this will "
                    "be used to create the `q_mu` and `q_sqrt` parameters."
                )

            num_inducing, self.num_latent_gps = (
                inducing_variable.num_inducing,
                num_latent_gps,
            )

        self.q_mu = Parameter(
            np.zeros((num_inducing, self.num_latent_gps)),
            dtype=default_float(),
            name=f"{self.name}_q_mu" if self.name else "q_mu",
        )  # [num_inducing, num_latent_gps]

        self.q_sqrt = Parameter(
            np.stack([np.eye(num_inducing) for _ in range(self.num_latent_gps)]),
            transform=triangular(),
            dtype=default_float(),
            name=f"{self.name}_q_sqrt" if self.name else "q_sqrt",
        )  # [num_latent_gps, num_inducing, num_inducing]

        self.num_samples = num_samples

    def predict(
        self,
        inputs: TensorType,
        *,
        full_cov: bool = False,
        full_output_cov: bool = False,
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Make a prediction at N test inputs for the Q outputs of this layer,
        including the mean function contribution.

        The covariance and its shape is determined by *full_cov* and *full_output_cov* as follows:

        +--------------------+---------------------------+--------------------------+
        | (co)variance shape | ``full_output_cov=False`` | ``full_output_cov=True`` |
        +--------------------+---------------------------+--------------------------+
        | ``full_cov=False`` | [N, Q]                    | [N, Q, Q]                |
        +--------------------+---------------------------+--------------------------+
        | ``full_cov=True``  | [Q, N, N]                 | [N, Q, N, Q]             |
        +--------------------+---------------------------+--------------------------+

        :param inputs: The inputs to predict at, with a shape of [N, D], where D is
            the input dimensionality of this layer.
        :param full_cov: Whether to return full covariance (if `True`) or
            marginal variance (if `False`, the default) w.r.t. inputs.
        :param full_output_cov: Whether to return full covariance (if `True`)
            or marginal variance (if `False`, the default) w.r.t. outputs.

        :returns: posterior mean (shape [N, Q]) and (co)variance (shape as above) at test points
        """
        mean_function = self.mean_function(inputs)
        mean_cond, cov = conditional(
            inputs,
            self.inducing_variable,
            self.kernel,
            self.q_mu,
            q_sqrt=self.q_sqrt,
            full_cov=full_cov,
            full_output_cov=full_output_cov,
            white=self.whiten,
        )

        return mean_cond + mean_function, cov

    def call(self, inputs: TensorType, *args: List[Any], **kwargs: Dict[str, Any]) -> tf.Tensor:
        """
        The default behaviour upon calling this layer.

        This method calls the `tfp.layers.DistributionLambda` super-class
        `call` method, which constructs a `tfp.distributions.Distribution`
        for the predictive distributions at the input points
        (see :meth:`_make_distribution_fn`).
        You can pass this distribution to `tf.convert_to_tensor`, which will return
        samples from the distribution (see :meth:`_convert_to_tensor_fn`).

        This method also adds a layer-specific loss function, given by the KL divergence between
        this layer and the GP prior (scaled to per-datapoint).
        """
        outputs = super().call(inputs, *args, **kwargs)

        if kwargs.get("training"):
            log_prior = tf.add_n([p.log_prior_density() for p in self.kernel.trainable_parameters])
            loss = self.prior_kl() - log_prior
            loss_per_datapoint = loss / self.num_data

        else:
            # TF quirk: add_loss must always add a tensor to compile
            loss_per_datapoint = tf.constant(0.0, dtype=default_float())
        self.add_loss(loss_per_datapoint)

        # Metric names should be unique; otherwise they get overwritten if you
        # have multiple with the same name
        name = f"{self.name}_prior_kl" if self.name else "prior_kl"
        self.add_metric(loss_per_datapoint, name=name, aggregation="mean")

        return outputs

    def prior_kl(self) -> tf.Tensor:
        r"""
        Returns the KL divergence ``KL[q(u)∥p(u)]`` from the prior ``p(u)`` to
        the variational distribution ``q(u)``.  If this layer uses the
        :attr:`whiten`\ ed representation, returns ``KL[q(v)∥p(v)]``.
        """
        return prior_kl(
            self.inducing_variable, self.kernel, self.q_mu, self.q_sqrt, whiten=self.whiten
        )

    def _make_distribution_fn(
        self, previous_layer_outputs: TensorType
    ) -> tfp.distributions.Distribution:
        """
        Construct the posterior distributions at the output points of the previous layer,
        depending on :attr:`full_cov` and :attr:`full_output_cov`.

        :param previous_layer_outputs: The output from the previous layer,
            which should be coercible to a `tf.Tensor`
        """
        mean, cov = self.predict(
            previous_layer_outputs,
            full_cov=self.full_cov,
            full_output_cov=self.full_output_cov,
        )

        if self.full_cov and not self.full_output_cov:
            # mean: [N, Q], cov: [Q, N, N]
            return tfp.distributions.MultivariateNormalTriL(
                loc=tf.linalg.adjoint(mean), scale_tril=_cholesky_with_jitter(cov)
            )  # loc: [Q, N], scale: [Q, N, N]
        elif self.full_output_cov and not self.full_cov:
            # mean: [N, Q], cov: [N, Q, Q]
            return tfp.distributions.MultivariateNormalTriL(
                loc=mean, scale_tril=_cholesky_with_jitter(cov)
            )  # loc: [N, Q], scale: [N, Q, Q]
        elif not self.full_cov and not self.full_output_cov:
            # mean: [N, Q], cov: [N, Q]
            return tfp.distributions.MultivariateNormalDiag(loc=mean, scale_diag=tf.sqrt(cov))
        else:
            raise NotImplementedError(
                "The combination of both `full_cov` and `full_output_cov` is not permitted."
            )

    def _convert_to_tensor_fn(self, distribution: tfp.distributions.Distribution) -> tf.Tensor:
        """
        Convert the predictive distributions at the input points (see
        :meth:`_make_distribution_fn`) to a tensor of :attr:`num_samples`
        samples from that distribution.
        Whether the samples are correlated or marginal (uncorrelated) depends
        on :attr:`full_cov` and :attr:`full_output_cov`.
        """
        # N input points
        # S = self.num_samples
        # Q = output dimensionality
        if self.num_samples is not None:
            samples = distribution.sample(
                (self.num_samples,)
            )  # [S, Q, N] if full_cov else [S, N, Q]
        else:
            samples = distribution.sample()  # [Q, N] if full_cov else [N, Q]

        if self.full_cov:
            samples = tf.linalg.adjoint(samples)  # [S, N, Q] or [N, Q]

        return samples

    def sample(self) -> Sample:
        """
        .. todo:: TODO: Document this.
        """
        return (
            efficient_sample(
                self.inducing_variable,
                self.kernel,
                self.q_mu,
                q_sqrt=self.q_sqrt,
                whiten=self.whiten,
            )
            # Makes use of the magic __add__ of the Sample class
            + self.mean_function
        )



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" This module contains a set of utilities used in experiments and notebooks. """

import os


def is_continuous_integration() -> bool:
    """
    Return `True` if the code is running on continuous integration (CI) machines,
    otherwise `False`.

    ..note:: This check is based on the ``CI`` environment variable, which is set to ``True``
        by GitHub actions, CircleCI, and Jenkins. This function may not work as expected
        under other CI frameworks.
    """
    ci = os.environ.get("CI", "").lower()
    return (ci == "true") or (ci == "1")


def notebook_niter(n: int, test_n: int = 2) -> int:
    """
    Return a typically smaller number of iterations ``test_n`` if
    code is running on CI machines (see :func:`is_continuous_integration`),
    otherwise return ``n``.
    """
    return test_n if is_continuous_integration() else n


def notebook_range(n: int, test_n: int = 2) -> range:
    """
    Return a typically shorter `range` (of the length specified by ``test_n``) if
    code is running on CI machines (see :func:`is_continuous_integration`),
    otherwise return a `range` of the length specified by ``n``.
    """
    return range(notebook_niter(n, test_n))


def notebook_list(lst: list, test_n: int = 2) -> list:
    """
    Return a subset of the length specified by ``test_n`` from a list ``lst`` if the code
    is running on CI machines (see :func:`is_continuous_integration`),
    otherwise return the complete list ``lst``.
    """
    return lst[:test_n] if is_continuous_integration() else lst



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Shared functionality for stationary kernel basis functions. """

from abc import ABC, abstractmethod
from typing import Mapping

import tensorflow as tf

import gpflow
from gpflow.base import TensorType

from gpflux.types import ShapeType


class FourierFeaturesBase(ABC, tf.keras.layers.Layer):
    def __init__(self, kernel: gpflow.kernels.Kernel, n_components: int, **kwargs: Mapping):
        """
        :param kernel: kernel to approximate using a set of Fourier bases.
        :param n_components: number of components (e.g. Monte Carlo samples,
            quadrature nodes, etc.) used to numerically approximate the kernel.
        """
        super(FourierFeaturesBase, self).__init__(**kwargs)
        self.kernel = kernel
        self.n_components = n_components
        if kwargs.get("input_dim", None):
            self._input_dim = kwargs["input_dim"]
            self.build(tf.TensorShape([self._input_dim]))
        else:
            self._input_dim = None

    def call(self, inputs: TensorType) -> tf.Tensor:
        """
        Evaluate the basis functions at ``inputs``.

        :param inputs: The evaluation points, a tensor with the shape ``[N, D]``.

        :return: A tensor with the shape ``[N, M]``.
        """
        X = tf.divide(inputs, self.kernel.lengthscales)  # [N, D]
        const = self._compute_constant()
        bases = self._compute_bases(X)
        output = const * bases
        tf.ensure_shape(output, self.compute_output_shape(inputs.shape))
        return output

    def compute_output_shape(self, input_shape: ShapeType) -> tf.TensorShape:
        """
        Computes the output shape of the layer.
        See `tf.keras.layers.Layer.compute_output_shape()
        <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#compute_output_shape>`_.
        """
        # TODO: Keras docs say "If the layer has not been built, this method
        # will call `build` on the layer." -- do we need to do so?
        tensor_shape = tf.TensorShape(input_shape).with_rank(2)
        output_dim = self._compute_output_dim(input_shape)
        return tensor_shape[:-1].concatenate(output_dim)

    def get_config(self) -> Mapping:
        """
        Returns the config of the layer.
        See `tf.keras.layers.Layer.get_config()
        <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#get_config>`_.
        """
        config = super(FourierFeaturesBase, self).get_config()
        config.update(
            {"kernel": self.kernel, "n_components": self.n_components, "input_dim": self._input_dim}
        )

        return config

    @abstractmethod
    def _compute_output_dim(self, input_shape: ShapeType) -> int:
        pass

    @abstractmethod
    def _compute_constant(self) -> tf.Tensor:
        """
        Compute normalizing constant for basis functions.
        """
        pass

    @abstractmethod
    def _compute_bases(self, inputs: TensorType) -> tf.Tensor:
        """
        Compute basis functions.
        """
        pass



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Mapping, Optional, Tuple, Type

import numpy as np
import tensorflow as tf

import gpflow
from gpflow.base import DType, TensorType

from gpflux.layers.basis_functions.fourier_features.base import FourierFeaturesBase
from gpflux.layers.basis_functions.fourier_features.utils import (
    _bases_concat,
    _bases_cosine,
    _matern_number,
)
from gpflux.types import ShapeType

"""
Kernels supported by :class:`RandomFourierFeatures`.

You can build RFF for shift-invariant stationary kernels from which you can
sample frequencies from their power spectrum, following Bochner's theorem.
"""
RFF_SUPPORTED_KERNELS: Tuple[Type[gpflow.kernels.Stationary], ...] = (
    gpflow.kernels.SquaredExponential,
    gpflow.kernels.Matern12,
    gpflow.kernels.Matern32,
    gpflow.kernels.Matern52,
)


def _sample_students_t(nu: float, shape: ShapeType, dtype: DType) -> TensorType:
    """
    Draw samples from a (central) Student's t-distribution using the following:
      BETA ~ Gamma(nu/2, nu/2) (shape-rate parameterization)
      X ~ Normal(0, 1/BETA)
    then:
      X ~ StudentsT(nu)

    Note this is equivalent to the more commonly used parameterization
      Z ~ Chi2(nu) = Gamma(nu/2, 1/2)
      EPSILON ~ Normal(0, 1)
      X = EPSILON * sqrt(nu/Z)

    To see this, note
      Z/nu ~ Gamma(nu/2, nu/2)
    and
      X ~ Normal(0, nu/Z)
    The equivalence becomes obvious when we set BETA = Z/nu
    """
    # Normal(0, 1)
    normal_rvs = tf.random.normal(shape=shape, dtype=dtype)
    shape = tf.concat([shape[:-1], [1]], axis=0)
    # Gamma(nu/2, nu/2)
    gamma_rvs = tf.random.gamma(shape, alpha=0.5 * nu, beta=0.5 * nu, dtype=dtype)
    # StudentsT(nu)
    students_t_rvs = tf.math.rsqrt(gamma_rvs) * normal_rvs
    return students_t_rvs


class RandomFourierFeaturesBase(FourierFeaturesBase):
    def __init__(self, kernel: gpflow.kernels.Kernel, n_components: int, **kwargs: Mapping):
        assert isinstance(kernel, RFF_SUPPORTED_KERNELS), "Unsupported Kernel"
        super(RandomFourierFeaturesBase, self).__init__(kernel, n_components, **kwargs)

    def build(self, input_shape: ShapeType) -> None:
        """
        Creates the variables of the layer.
        See `tf.keras.layers.Layer.build()
        <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#build>`_.
        """
        input_dim = input_shape[-1]
        self._weights_build(input_dim, n_components=self.n_components)
        super(RandomFourierFeaturesBase, self).build(input_shape)

    def _weights_build(self, input_dim: int, n_components: int) -> None:
        shape = (n_components, input_dim)
        self.W = self.add_weight(
            name="weights",
            trainable=False,
            shape=shape,
            dtype=self.dtype,
            initializer=self._weights_init,
        )

    def _weights_init(self, shape: TensorType, dtype: Optional[DType] = None) -> TensorType:
        if isinstance(self.kernel, gpflow.kernels.SquaredExponential):
            return tf.random.normal(shape, dtype=dtype)
        else:
            p = _matern_number(self.kernel)
            nu = 2.0 * p + 1.0  # degrees of freedom
            return _sample_students_t(nu, shape, dtype)

    @staticmethod
    def rff_constant(variance: TensorType, output_dim: int) -> tf.Tensor:
        """
        Normalizing constant for random Fourier features.
        """
        return tf.sqrt(tf.math.truediv(2.0 * variance, output_dim))


class RandomFourierFeatures(RandomFourierFeaturesBase):
    r"""
    Random Fourier features (RFF) is a method for approximating kernels. The essential
    element of the RFF approach :cite:p:`rahimi2007random` is the realization that Bochner's theorem
    for stationary kernels can be approximated by a Monte Carlo sum.

    We will approximate the kernel :math:`k(\mathbf{x}, \mathbf{x}')`
    by :math:`\Phi(\mathbf{x})^\top \Phi(\mathbf{x}')`
    where :math:`\Phi: \mathbb{R}^{D} \to \mathbb{R}^{M}` is a finite-dimensional feature map.

    The feature map is defined as:

    .. math::

      \Phi(\mathbf{x}) = \sqrt{\frac{2 \sigma^2}{\ell}}
        \begin{bmatrix}
          \cos(\boldsymbol{\theta}_1^\top \mathbf{x}) \\
          \sin(\boldsymbol{\theta}_1^\top \mathbf{x}) \\
          \vdots \\
          \cos(\boldsymbol{\theta}_{\frac{M}{2}}^\top \mathbf{x}) \\
          \sin(\boldsymbol{\theta}_{\frac{M}{2}}^\top \mathbf{x})
        \end{bmatrix}

    where :math:`\sigma^2` is the kernel variance.
    The features are parameterised by random weights:

    - :math:`\boldsymbol{\theta} \sim p(\boldsymbol{\theta})`
      where :math:`p(\boldsymbol{\theta})` is the spectral density of the kernel.

    At least for the squared exponential kernel, this variant of the feature
    mapping has more desirable theoretical properties than its counterpart form
    from phase-shifted cosines :class:`RandomFourierFeaturesCosine` :cite:p:`sutherland2015error`.
    """

    def _compute_output_dim(self, input_shape: ShapeType) -> int:
        return 2 * self.n_components

    def _compute_bases(self, inputs: TensorType) -> tf.Tensor:
        """
        Compute basis functions.

        :return: A tensor with the shape ``[N, 2M]``.
        """
        return _bases_concat(inputs, self.W)

    def _compute_constant(self) -> tf.Tensor:
        """
        Compute normalizing constant for basis functions.

        :return: A tensor with the shape ``[]`` (i.e. a scalar).
        """
        return self.rff_constant(self.kernel.variance, output_dim=2 * self.n_components)


class RandomFourierFeaturesCosine(RandomFourierFeaturesBase):
    r"""
    Random Fourier Features (RFF) is a method for approximating kernels. The essential
    element of the RFF approach :cite:p:`rahimi2007random` is the realization that Bochner's theorem
    for stationary kernels can be approximated by a Monte Carlo sum.

    We will approximate the kernel :math:`k(\mathbf{x}, \mathbf{x}')`
    by :math:`\Phi(\mathbf{x})^\top \Phi(\mathbf{x}')` where
    :math:`\Phi: \mathbb{R}^{D} \to \mathbb{R}^{M}` is a finite-dimensional feature map.

    The feature map is defined as:

    .. math::
      \Phi(\mathbf{x}) = \sqrt{\frac{2 \sigma^2}{\ell}}
        \begin{bmatrix}
          \cos(\boldsymbol{\theta}_1^\top \mathbf{x} + \tau) \\
          \vdots \\
          \cos(\boldsymbol{\theta}_M^\top \mathbf{x} + \tau)
        \end{bmatrix}

    where :math:`\sigma^2` is the kernel variance.
    The features are parameterised by random weights:

    - :math:`\boldsymbol{\theta} \sim p(\boldsymbol{\theta})`
      where :math:`p(\boldsymbol{\theta})` is the spectral density of the kernel
    - :math:`\tau \sim \mathcal{U}(0, 2\pi)`

    Equivalent to :class:`RandomFourierFeatures` by elementary trigonometric identities.
    """

    def build(self, input_shape: ShapeType) -> None:
        """
        Creates the variables of the layer.
        See `tf.keras.layers.Layer.build()
        <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Layer#build>`_.
        """
        self._bias_build(n_components=self.n_components)
        super(RandomFourierFeaturesCosine, self).build(input_shape)

    def _bias_build(self, n_components: int) -> None:
        shape = (1, n_components)
        self.b = self.add_weight(
            name="bias",
            trainable=False,
            shape=shape,
            dtype=self.dtype,
            initializer=self._bias_init,
        )

    def _bias_init(self, shape: TensorType, dtype: Optional[DType] = None) -> TensorType:
        return tf.random.uniform(shape=shape, maxval=2.0 * np.pi, dtype=dtype)

    def _compute_output_dim(self, input_shape: ShapeType) -> int:
        return self.n_components

    def _compute_bases(self, inputs: TensorType) -> tf.Tensor:
        """
        Compute basis functions.

        :return: A tensor with the shape ``[N, M]``.
        """
        return _bases_cosine(inputs, self.W, self.b)

    def _compute_constant(self) -> tf.Tensor:
        """
        Compute normalizing constant for basis functions.

        :return: A tensor with the shape ``[]`` (i.e. a scalar).
        """
        return self.rff_constant(self.kernel.variance, output_dim=self.n_components)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility layer that tracks variables in :class:`tf.Module`."""

import tensorflow as tf
from deprecated import deprecated


@deprecated(
    reason=(
        "GPflux's `TrackableLayer` was prior to TF2.5 used to collect GPflow "
        "variables in subclassed layers. As of TF 2.5, `tf.Module` supports "
        "this natively and there is no need for `TrackableLayer` anymore. It will "
        "be removed in GPflux version `1.0.0`."
    )
)
class TrackableLayer(tf.keras.layers.Layer):
    """
    With the release of TensorFlow 2.5, our TrackableLayer workaround is no
    longer needed.  See https://github.com/Prowler-io/gpflux/issues/189.
    Will be removed in GPflux version 1.0.0
    """

    pass



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
r"""
This module implements a callback that enables GPflow's `gpflow.monitor.ModelToTensorBoard` to
integrate with Keras's `tf.keras.Model`\ 's `fit
<https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit>`_ method.
"""
import re
from typing import Any, Dict, List, Mapping, Optional, Union

import tensorflow as tf

import gpflow
from gpflow.utilities import parameter_dict

__all__ = ["TensorBoard"]


class TensorBoard(tf.keras.callbacks.TensorBoard):
    """
    This class is a thin wrapper around a `tf.keras.callbacks.TensorBoard` callback that also
    calls GPflow's `gpflow.monitor.ModelToTensorBoard` monitoring task.
    """

    log_dir: str
    """
    The path of the directory to which to save the log files to be
    read by TensorBoard. Files are saved in ``log_dir / "train"``,
    following the Keras convention.
    """

    keywords_to_monitor: List[str]
    """
    Parameters whose name match a keyword in the *keywords_to_monitor* list
    will be added to the TensorBoard.
    """

    update_freq: Union[int, str]
    r"""
    Either an integer or ``"epoch"``. If using an integer *n*, write
    losses/metrics/parameters at every *n*\ th batch; when using
    ``"epoch"``, write them at the end of each epoch. Note that writing too
    frequently to TensorBoard can slow down the training.
    """

    def __init__(
        self,
        log_dir: str = "logs",
        *,
        keywords_to_monitor: List[str] = [
            "kernel",
            "mean_function",
            "likelihood",
        ],
        max_size: int = 3,
        histogram_freq: int = 0,
        write_graph: bool = True,
        write_images: bool = False,
        update_freq: Union[int, str] = "epoch",
        profile_batch: int = 2,
        embeddings_freq: int = 0,
        embeddings_metadata: Optional[Dict] = None,
    ):
        """
        :param log_dir: The path of the directory to which to save the log files to be
            read by TensorBoard.
        :param keywords_to_monitor: A list of keywords to monitor.
            If the parameter's name includes any of the keywords specified, it
            will be added to the TensorBoard.
        :param max_size: The maximum size of arrays (inclusive) for which each element is
            written independently as a scalar to the TensorBoard.

        For information on all other arguments, see TensorFlow's `TensorBoard callback docs
        <https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/TensorBoard>`_.
        """

        super().__init__(
            log_dir=log_dir,
            histogram_freq=histogram_freq,
            write_graph=write_graph,
            write_images=write_images,
            update_freq=update_freq,
            profile_batch=profile_batch,
            embeddings_freq=embeddings_freq,
            embeddings_metadata=embeddings_metadata,
        )
        self.keywords_to_monitor = keywords_to_monitor
        self.max_size = max_size

    def set_model(self, model: tf.keras.Model) -> None:
        """
        Set the model (extends the Keras `set_model
        <https://www.tensorflow.org/api_docs/python/tf/keras/callbacks/TensorBoard#set_model>`_
        method).

        This method initialises :class:`KerasModelToTensorBoard` and mimics Keras's TensorBoard
        callback in writing the summary logs to :attr:`log_dir` / "train".
        """
        super().set_model(model)
        self.monitor = KerasModelToTensorBoard(
            log_dir=self.log_dir + "/train",
            model=self.model,
            max_size=self.max_size,
            keywords_to_monitor=list(self.keywords_to_monitor),
            left_strip_character="._",
        )

    def on_train_batch_end(self, batch: int, logs: Optional[Mapping] = None) -> None:
        """
        Write to TensorBoard if :attr:`update_freq` is an integer.

        :param batch: The index of the batch within the current epoch.
        :param logs: The metric results for this batch.
        """
        super().on_train_batch_end(batch, logs=logs)

        if self.update_freq == "epoch":
            # We only write at epoch ends
            return

        if isinstance(self.update_freq, int) and batch % self.update_freq == 0:
            self.monitor(batch)

    def on_epoch_end(self, epoch: int, logs: Optional[Mapping] = None) -> None:
        """Write to TensorBoard if :attr:`update_freq` equals ``"epoch"``."""
        super().on_epoch_end(epoch, logs=logs)

        if self.update_freq == "epoch":
            self.monitor(epoch)


class KerasModelToTensorBoard(gpflow.monitor.ModelToTensorBoard):
    """
    This class overwrites the :meth:`run` method to deduplicate parameters in
    :attr:`parameter_dict`.
    """

    # Keras automatically saves all layers in the `self._self_tracked_trackables`
    # attribute of the model, which is a list of sub-modules.
    _LAYER_PARAMETER_REGEXP = re.compile(r"\._self_tracked_trackables\[\d+\]\.")

    def _parameter_of_interest(self, match: str) -> bool:
        return self._LAYER_PARAMETER_REGEXP.match(match) is not None

    def run(self, **unused_kwargs: Any) -> None:
        """Write the model's parameters to TensorBoard."""

        for name, parameter in parameter_dict(self.model).items():
            if not self._parameter_of_interest(name):
                # skip parameters
                continue
            # check if the parameter name matches any of the specified keywords
            if self.summarize_all or any((keyword in name) for keyword in self.keywords_to_monitor):
                # keys are sometimes prepended with a character, which we strip
                name = name.lstrip(self.left_strip_character)
                self._summarize_parameter(name, parameter)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Types used within GPflux (for static type-checking).
"""
from typing import List, Tuple, Union

import tensorflow as tf
import tensorflow_probability as tfp

from gpflow.base import TensorType


def unwrap_dist(dist: tfp.distributions.Distribution) -> tfp.distributions.Distribution:
    """
    Unwrap the given distribution, if it is wrapped in a ``_TensorCoercible``.
    """
    while True:
        inner = getattr(dist, "tensor_distribution", None)
        if inner is None:
            return dist
        dist = inner


ShapeType = Union[tf.TensorShape, List[int], Tuple[int, ...]]
r""" Union of valid types for describing the shape of a `tf.Tensor`\ (-like) object """

ObservationType = List[TensorType]
"""
Type for the ``[inputs, targets]`` list used by :class:`~gpflux.layers.LayerWithObservations`
"""



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides :class:`BayesianDenseLayer`, which implements a
variational Bayesian dense (fully-connected) neural network layer as a Keras
:class:`~tf.keras.layers.Layer`.
"""

from typing import Callable, Optional, Union

import numpy as np
import tensorflow as tf

from gpflow import Parameter, default_float
from gpflow.base import TensorType
from gpflow.kullback_leiblers import gauss_kl
from gpflow.models.model import MeanAndVariance
from gpflow.utilities.bijectors import positive, triangular

from gpflux.helpers import xavier_initialization_numpy
from gpflux.layers.trackable_layer import TrackableLayer
from gpflux.types import ShapeType


class BayesianDenseLayer(TrackableLayer):
    """
    A dense (fully-connected) layer for variational Bayesian neural networks.

    This layer holds the mean and square-root of the variance of the
    distribution over the weights. This layer also has a temperature for
    cooling (or heating) the posterior.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_data: int,
        w_mu: Optional[np.ndarray] = None,
        w_sqrt: Optional[np.ndarray] = None,
        activation: Optional[Callable] = None,
        is_mean_field: bool = True,
        temperature: float = 1e-4,
    ):
        """
        :param input_dim: The input dimension (excluding bias) of this layer.
        :param output_dim: The output dimension of this layer.
        :param num_data: The number of points in the training dataset (used for
            scaling the KL regulariser).
        :param w_mu: Initial value of the variational mean for weights + bias.
            If not specified, this defaults to `xavier_initialization_numpy`
            for the weights and zero for the bias.
        :param w_sqrt: Initial value of the variational Cholesky of the
            (co)variance for weights + bias. If not specified, this defaults to
            1e-5 * Identity.
        :param activation: The activation function. If not specified, this defaults to the identity.
        :param is_mean_field: Determines whether the approximation to the
            weight posterior is mean field. Must be consistent with the shape
            of ``w_sqrt``, if specified.
        :param temperature: The KL loss will be scaled by this factor.
            Can be used for cooling (< 1.0) or heating (> 1.0) the posterior.
            As suggested in `"How Good is the Bayes Posterior in Deep Neural
            Networks Really?" by Wenzel et al. (2020)
            <http://proceedings.mlr.press/v119/wenzel20a>`_ the default value
            is a cold ``1e-4``.
        """

        super().__init__(dtype=default_float())

        assert input_dim >= 1
        assert output_dim >= 1
        assert num_data >= 1

        if w_mu is not None:  # add + 1 for the bias
            assert w_mu.shape == ((input_dim + 1) * output_dim,)
        if w_sqrt is not None:
            if not is_mean_field:
                assert w_sqrt.shape == (
                    (input_dim + 1) * output_dim,
                    (input_dim + 1) * output_dim,
                )
            else:
                assert w_sqrt.shape == ((input_dim + 1) * output_dim,)
        assert temperature > 0.0

        self.input_dim = input_dim
        self.output_dim = output_dim
        self.num_data = num_data

        self.w_mu_ini = w_mu
        self.w_sqrt_ini = w_sqrt

        self.activation = activation
        self.is_mean_field = is_mean_field
        self.temperature = temperature

        self.dim = (input_dim + 1) * output_dim
        self.full_output_cov = False
        self.full_cov = False

        self.w_mu = Parameter(np.zeros((self.dim,)), dtype=default_float(), name="w_mu")  # [dim]

        self.w_sqrt = Parameter(
            np.zeros((self.dim, self.dim)) if not self.is_mean_field else np.ones((self.dim,)),
            transform=triangular() if not self.is_mean_field else positive(),
            dtype=default_float(),
            name="w_sqrt",
        )  # [dim, dim] or [dim]

    def initialize_variational_distribution(self) -> None:
        if self.w_mu_ini is None:
            w = xavier_initialization_numpy(self.input_dim, self.output_dim)
            b = np.zeros((1, self.output_dim))
            self.w_mu_ini = np.concatenate((w, b), axis=0).reshape((self.dim,))
        self.w_mu.assign(self.w_mu_ini)

        if self.w_sqrt_ini is None:
            if not self.is_mean_field:
                self.w_sqrt_ini = 1e-5 * np.eye(self.dim)
            else:
                self.w_sqrt_ini = 1e-5 * np.ones((self.dim,))
        self.w_sqrt.assign(self.w_sqrt_ini)

    def build(self, input_shape: ShapeType) -> None:
        """Build the variables necessary on first call"""
        super().build(input_shape)
        self.initialize_variational_distribution()

    def predict_samples(
        self,
        inputs: TensorType,
        *,
        num_samples: Optional[int] = None,
    ) -> tf.Tensor:
        """
        Samples from the approximate posterior at N test inputs, with input_dim = D, output_dim = Q.

        :param inputs: The inputs to predict at; shape ``[N, D]``.
        :param num_samples: The number of samples S, to draw.
        :returns: Samples, shape ``[S, N, Q]`` if S is not None else ``[N, Q]``.
        """
        _num_samples = num_samples or 1
        z = tf.random.normal((self.dim, _num_samples), dtype=default_float())  # [dim, S]
        if not self.is_mean_field:
            w = self.w_mu[:, None] + tf.matmul(self.w_sqrt, z)  # [dim, S]
        else:
            w = self.w_mu[:, None] + self.w_sqrt[:, None] * z  # [dim, S]

        N = tf.shape(inputs)[0]
        inputs_concat_1 = tf.concat(
            (inputs, tf.ones((N, 1), dtype=default_float())), axis=-1
        )  # [N, D+1]
        samples = tf.tensordot(
            inputs_concat_1,
            tf.reshape(tf.transpose(w), (_num_samples, self.input_dim + 1, self.output_dim)),
            [[-1], [1]],
        )  # [N, S, Q]
        if num_samples is None:
            samples = tf.squeeze(samples, axis=-2)  # [N, Q]
        else:
            samples = tf.transpose(samples, perm=[1, 0, 2])  # [S, N, Q]

        if self.activation is not None:
            samples = self.activation(samples)

        return samples

    def call(
        self, inputs: TensorType, training: Optional[bool] = False
    ) -> Union[tf.Tensor, MeanAndVariance]:
        """
        The default behaviour upon calling this layer.
        """
        sample = self.predict_samples(
            inputs,
            num_samples=None,
        )

        # TF quirk: add_loss must add a tensor to compile
        if training:
            loss = self.temperature * self.prior_kl()
        else:
            loss = tf.constant(0.0, dtype=default_float())
        loss_per_datapoint = loss / self.num_data

        self.add_loss(loss_per_datapoint)

        return sample  # [N, Q]

    def prior_kl(self) -> tf.Tensor:
        """
        Returns the KL divergence ``KL[q(u)∥p(u)]`` from the prior ``p(u) = N(0, I)`` to
        the variational distribution ``q(u) = N(w_mu, w_sqrt²)``.
        """
        return gauss_kl(
            self.w_mu[:, None],
            self.w_sqrt[None] if not self.is_mean_field else self.w_sqrt[:, None],
        )



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
A Keras Layer that wraps a likelihood, while containing the necessary operations
for training.
"""
from typing import Optional

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow_probability.python.util.deferred_tensor import TensorMetaClass

from gpflow import default_float
from gpflow.base import TensorType
from gpflow.likelihoods import Likelihood

from gpflux.layers.trackable_layer import TrackableLayer
from gpflux.types import unwrap_dist


class LikelihoodLayer(TrackableLayer):
    r"""
    A Keras layer that wraps a GPflow :class:`~gpflow.likelihoods.Likelihood`. This layer expects a
    `tfp.distributions.MultivariateNormalDiag` as its input, describing ``q(f)``.
    When training, calling this class computes the negative variational expectation
    :math:`-\mathbb{E}_{q(f)}[\log p(y|f)]` and adds it as a layer loss.
    When not training, it computes the mean and variance of ``y`` under ``q(f)``
    using :meth:`~gpflow.likelihoods.Likelihood.predict_mean_and_var`.

    .. note::

        Use **either** this `LikelihoodLayer` (together with
        `gpflux.models.DeepGP`) **or** `LikelihoodLoss` (e.g. together with a
        `tf.keras.Sequential` model). Do **not** use both at once because
        this would add the loss twice.
    """

    def __init__(self, likelihood: Likelihood):
        super().__init__(dtype=default_float())
        self.likelihood = likelihood

    def call(
        self,
        inputs: tfp.distributions.MultivariateNormalDiag,
        targets: Optional[TensorType] = None,
        training: bool = None,
    ) -> "LikelihoodOutputs":
        """
        When training (``training=True``), this method computes variational expectations
        (data-fit loss) and adds this information as a layer loss.
        When testing (the default), it computes the posterior mean and variance of ``y``.

        :param inputs: The output distribution of the previous layer. This is currently
            expected to be a :class:`~tfp.distributions.MultivariateNormalDiag`;
            that is, the preceding :class:`~gpflux.layers.GPLayer` should have
            ``full_cov=full_output_cov=False``.
        :returns: a `LikelihoodOutputs` tuple with the mean and variance of ``f`` and,
            if not training, the mean and variance of ``y``.

        .. todo:: Turn this layer into a
            :class:`~tfp.layers.DistributionLambda` as well and return the
            correct :class:`~tfp.distributions.Distribution` instead of a tuple
            containing mean and variance only.
        """
        # TODO: add support for non-distribution inputs? or other distributions?
        assert isinstance(unwrap_dist(inputs), tfp.distributions.MultivariateNormalDiag)
        no_X = None
        F_mean = inputs.loc
        F_var = inputs.scale.diag ** 2

        if training:
            assert targets is not None
            # TODO: re-use LikelihoodLoss to remove code duplication
            loss_per_datapoint = tf.reduce_mean(
                -self.likelihood.variational_expectations(no_X, F_mean, F_var, targets)
            )
            Y_mean = Y_var = None
        else:
            loss_per_datapoint = tf.constant(0.0, dtype=default_float())
            Y_mean, Y_var = self.likelihood.predict_mean_and_var(no_X, F_mean, F_var)

        self.add_loss(loss_per_datapoint)

        # TODO: turn this layer into a DistributionLambda as well and
        # return correct distribution, not a mean/variance tuple
        return LikelihoodOutputs(F_mean, F_var, Y_mean, Y_var)


class LikelihoodOutputs(tf.Module, metaclass=TensorMetaClass):
    """
    This class encapsulates the outputs of a :class:`~gpflux.layers.LikelihoodLayer`.

    It contains the mean and variance of the marginal distribution of the final latent
    :class:`~gpflux.layers.GPLayer`, as well as the mean and variance of the likelihood.

    This class includes the `TensorMetaClass
    <https://github.com/tensorflow/probability/blob/master/tensorflow_probability/python/util/deferred_tensor.py#L81>`_
    to make objects behave as a `tf.Tensor`. This is necessary so that it can be
    returned from the `tfp.layers.DistributionLambda` Keras layer.
    """

    def __init__(
        self,
        f_mean: TensorType,
        f_var: TensorType,
        y_mean: Optional[TensorType],
        y_var: Optional[TensorType],
    ):
        super().__init__(name="likelihood_outputs")

        self.f_mean = f_mean
        self.f_var = f_var
        self.y_mean = y_mean
        self.y_var = y_var

    def _value(
        self, dtype: tf.dtypes.DType = None, name: str = None, as_ref: bool = False
    ) -> tf.Tensor:
        return self.f_mean

    @property
    def shape(self) -> tf.Tensor:
        return self.f_mean.shape

    @property
    def dtype(self) -> tf.dtypes.DType:
        return self.f_mean.dtype



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
r"""
The classes in this module encapsulate kernels :math:`k(\cdot, \cdot)` with
their features :math:`\phi_i(\cdot)` and coefficients :math:`\lambda_i` so
that:

.. math::

    k(x, x') = \sum_{i=0}^\infty \lambda_i \phi_i(x) \phi_i(x').

The kernels are used for efficient sampling. See the tutorial notebooks
`Efficient sampling <../../../../notebooks/efficient_sampling.ipynb>`_
and `Weight Space Approximation with Random Fourier Features
<../../../../notebooks/weight_space_approximation.ipynb>`_
for an in-depth overview.
"""
from typing import Optional, Union

import tensorflow as tf

import gpflow
from gpflow.base import TensorType

NoneType = type(None)


class _ApproximateKernel(gpflow.kernels.Kernel):
    r"""
    This class approximates a kernel by the finite feature decomposition:

    .. math:: k(x, x') = \sum_{i=0}^L \lambda_i \phi_i(x) \phi_i(x'),

    where :math:`\lambda_i` and :math:`\phi_i(\cdot)` are the coefficients
    and features, respectively.

    """

    def __init__(
        self,
        feature_functions: tf.keras.layers.Layer,
        feature_coefficients: TensorType,
    ):
        r"""
        :param feature_functions: A Keras layer for which the call evaluates the
            ``L`` features of the kernel :math:`\phi_i(\cdot)`. For ``X`` with the shape ``[N, D]``,
            ``feature_functions(X)`` returns a tensor with the shape ``[N, L]``.
        :param feature_coefficients: A tensor with the shape ``[L, 1]`` with coefficients
            associated with the features, :math:`\lambda_i`.
        """
        self._feature_functions = feature_functions
        self._feature_coefficients = feature_coefficients  # [L, 1]

    def K(self, X: TensorType, X2: Optional[TensorType] = None) -> tf.Tensor:
        """Approximate the true kernel by an inner product between feature functions."""
        phi = self._feature_functions(X)  # [N, L]
        if X2 is None:
            phi2 = phi
        else:
            phi2 = self._feature_functions(X2)  # [N2, L]

        r = tf.matmul(
            phi, tf.transpose(self._feature_coefficients) * phi2, transpose_b=True
        )  # [N, N2]

        N1, N2 = tf.shape(phi)[0], tf.shape(phi2)[0]
        tf.debugging.assert_equal(tf.shape(r), [N1, N2])
        return r

    def K_diag(self, X: TensorType) -> tf.Tensor:
        """Approximate the true kernel by an inner product between feature functions."""
        phi_squared = self._feature_functions(X) ** 2  # [N, L]
        r = tf.reduce_sum(phi_squared * tf.transpose(self._feature_coefficients), axis=1)  # [N,]
        N = tf.shape(X)[0]
        tf.debugging.assert_equal(tf.shape(r), [N])  # noqa: E231
        return r


class KernelWithFeatureDecomposition(gpflow.kernels.Kernel):
    r"""
    This class represents a kernel together with its finite feature decomposition:

    .. math:: k(x, x') = \sum_{i=0}^L \lambda_i \phi_i(x) \phi_i(x'),

    where :math:`\lambda_i` and :math:`\phi_i(\cdot)` are the coefficients and
    features, respectively.

    The decomposition can be derived from Mercer or Bochner's theorem. For example,
    feature-coefficient pairs could be eigenfunction-eigenvalue pairs (Mercer) or
    Fourier features with constant coefficients (Bochner).

    In some cases (e.g., [1]_ and [2]_) the left-hand side (that is, the
    covariance function :math:`k(\cdot, \cdot)`) is unknown and the kernel
    can only be approximated using its feature decomposition.
    In other cases (e.g., [3]_ and [4]_), both the covariance function and feature
    decomposition are available in closed form.

    .. [1]
        Solin, Arno, and Simo Särkkä. "Hilbert space methods for
        reduced-rank Gaussian process regression." Statistics and Computing
        (2020).
    .. [2]
        Borovitskiy, Viacheslav, et al. "Matérn Gaussian processes on
        Riemannian manifolds." In Advances in Neural Information Processing
        Systems (2020).
    .. [3]
        Ali Rahimi and Benjamin Recht. Random features for large-scale kernel
        machines. In Advances in Neural Information Processing Systems (2007).
    .. [4]
        Dutordoir, Vincent, Nicolas Durrande, and James Hensman. "Sparse
        Gaussian processes with spherical harmonic features." In International
        Conference on Machine Learning (2020).
    """

    def __init__(
        self,
        kernel: Union[gpflow.kernels.Kernel, NoneType],
        feature_functions: tf.keras.layers.Layer,
        feature_coefficients: TensorType,
    ):
        r"""
        :param kernel: The kernel corresponding to the feature decomposition.
            If ``None``, there is no analytical expression associated with the infinite
            sum and we approximate the kernel based on the feature decomposition.

            .. note::

                In certain cases, the analytical expression for the kernel is
                not available. In this case, passing `None` is allowed, and
                :meth:`K` and :meth:`K_diag` will be computed using the
                approximation provided by the feature decomposition.

        :param feature_functions: A Keras layer for which the call evaluates the
            ``L`` features of the kernel :math:`\phi_i(\cdot)`. For ``X`` with the shape ``[N, D]``,
            ``feature_functions(X)`` returns a tensor with the shape ``[N, L]``.
        :param feature_coefficients: A tensor with the shape ``[L, 1]`` with coefficients
            associated with the features, :math:`\lambda_i`.
        """
        super().__init__()

        if kernel is None:
            self._kernel = _ApproximateKernel(feature_functions, feature_coefficients)
        else:
            self._kernel = kernel

        self._feature_functions = feature_functions
        self._feature_coefficients = feature_coefficients  # [L, 1]
        tf.ensure_shape(self._feature_coefficients, tf.TensorShape([None, 1]))

    @property
    def feature_functions(self) -> tf.keras.layers.Layer:
        r"""Return the kernel's features :math:`\phi_i(\cdot)`."""
        return self._feature_functions

    @property
    def feature_coefficients(self) -> tf.Tensor:
        r"""Return the kernel's coefficients :math:`\lambda_i`."""
        return self._feature_coefficients

    def K(self, X: TensorType, X2: Optional[TensorType] = None) -> tf.Tensor:
        return self._kernel.K(X, X2)

    def K_diag(self, X: TensorType) -> tf.Tensor:
        return self._kernel.K_diag(X)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
r"""
This module contains helper functions for constructing :class:`~gpflow.kernels.MultioutputKernel`,
:class:`~gpflow.inducing_variables.MultioutputInducingVariables`,
:class:`~gpflow.mean_functions.MeanFunction`, and :class:`~gpflux.layers.GPLayer` objects.
"""

import inspect
import warnings
from dataclasses import fields
from typing import List, Optional, Type, TypeVar, Union

import numpy as np

import gpflow
from gpflow import default_float
from gpflow.inducing_variables import (
    InducingPoints,
    SeparateIndependentInducingVariables,
    SharedIndependentInducingVariables,
)
from gpflow.kernels import SeparateIndependent, SharedIndependent
from gpflow.utilities import deepcopy

from gpflux.layers.gp_layer import GPLayer


def construct_basic_kernel(
    kernels: Union[gpflow.kernels.Kernel, List[gpflow.kernels.Kernel]],
    output_dim: Optional[int] = None,
    share_hyperparams: bool = False,
) -> gpflow.kernels.MultioutputKernel:
    r"""
    Construct a :class:`~gpflow.kernels.MultioutputKernel` to use
    in :class:`GPLayer`\ s.

    :param kernels: A single kernel or list of :class:`~gpflow.kernels.Kernel`\ s.
        - When a single kernel is passed, the same kernel is used for all
        outputs. Depending on ``share_hyperparams``, the hyperparameters will be
        shared across outputs. You must also specify ``output_dim``.
        - When a list of kernels is passed, each kernel in the list is used on a separate
        output dimension and a :class:`gpflow.kernels.SeparateIndependent` is returned.
    :param output_dim: The number of outputs. This is equal to the number of latent GPs
        in the :class:`GPLayer`. When only a single kernel is specified for ``kernels``,
        you must also specify ``output_dim``. When a list of kernels is specified for ``kernels``,
        we assume that ``len(kernels) == output_dim``, and ``output_dim`` is not required.
    :param share_hyperparams: If `True`, use the type of kernel and the same hyperparameters
        (variance and lengthscales) for the different outputs. Otherwise, the
        same type of kernel (Squared-Exponential, Matern12, and so on) is used for
        the different outputs, but the kernel can have different hyperparameter values for each.
    """
    if isinstance(kernels, list):
        mo_kern = SeparateIndependent(kernels)
    elif not share_hyperparams:
        copies = [deepcopy(kernels) for _ in range(output_dim)]
        mo_kern = SeparateIndependent(copies)
    else:
        mo_kern = SharedIndependent(kernels, output_dim)
    return mo_kern


def construct_basic_inducing_variables(
    num_inducing: Union[int, List[int]],
    input_dim: int,
    output_dim: Optional[int] = None,
    share_variables: bool = False,
    z_init: Optional[np.ndarray] = None,
) -> gpflow.inducing_variables.MultioutputInducingVariables:
    r"""
    Construct a compatible :class:`~gpflow.inducing_variables.MultioutputInducingVariables`
    to use in :class:`GPLayer`\ s.

    :param num_inducing: The total number of inducing variables, ``M``.
        This parameter can be freely chosen by the user. General advice
        is to set it as high as possible, but smaller than the number of datapoints.
        The computational complexity of the layer is cubic in ``M``.
        If a list is passed, each element in the list specifies the number of inducing
        variables to use for each ``output_dim``.
    :param input_dim: The dimensionality of the input data (or features) ``X``.
        Typically, this corresponds to ``X.shape[-1]``.
        For :class:`~gpflow.inducing_variables.InducingPoints`, this specifies the dimensionality
        of ``Z``.
    :param output_dim: The dimensionality of the outputs (or targets) ``Y``.
        Typically, this corresponds to ``Y.shape[-1]`` or the number of latent GPs.
        The parameter is used to determine the number of inducing variable sets
        to create when a different set is used for each output. The parameter
        is redundant when ``num_inducing`` is a list, because the code assumes
        that ``len(num_inducing) == output_dim``.
    :param share_variables: If `True`, use the same inducing variables for different
        outputs. Otherwise, create a different set for each output. Set this parameter to
        `False` when ``num_inducing`` is a list, because otherwise the two arguments
        contradict each other. If you set this parameter to `True`, you must also specify
        ``output_dim``, because that is used to determine the number of inducing variable
        sets to create.
    :param z_init: Raw values to use to initialise
        :class:`gpflow.inducing_variables.InducingPoints`. If `None` (the default), values
        will be initialised from ``N(0, 1)``. The shape of ``z_init`` depends on the other
        input arguments. If a single set of inducing points is used for all outputs (that
        is, if ``share_variables`` is `True`), ``z_init`` should be rank two, with the
        dimensions ``[M, input_dim]``. If a different set of inducing points is used for
        the outputs (ithat is, if ``num_inducing`` is a list, or if ``share_variables`` is
        `False`), ``z_init`` should be a rank three tensor with the dimensions
        ``[output_dim, M, input_dim]``.
    """

    if z_init is None:
        warnings.warn(
            "No `z_init` has been specified in `construct_basic_inducing_variables`. "
            "Default initialization using random normal N(0, 1) will be used."
        )

    z_init_is_given = z_init is not None

    if isinstance(num_inducing, list):
        if output_dim is not None:
            # TODO: the following assert may clash with MixedMultiOutputFeatures
            # where the number of independent GPs can differ from the output
            # dimension
            assert output_dim == len(num_inducing)  # pragma: no cover
        assert share_variables is False

        inducing_variables = []
        for i, num_ind_var in enumerate(num_inducing):
            if z_init_is_given:
                assert len(z_init[i]) == num_ind_var
                z_init_i = z_init[i]
            else:
                z_init_i = np.random.randn(num_ind_var, input_dim).astype(dtype=default_float())
            assert z_init_i.shape == (num_ind_var, input_dim)
            inducing_variables.append(InducingPoints(z_init_i))
        return SeparateIndependentInducingVariables(inducing_variables)

    elif not share_variables:
        inducing_variables = []
        for o in range(output_dim):
            if z_init_is_given:
                if z_init.shape != (output_dim, num_inducing, input_dim):
                    raise ValueError(
                        "When not sharing variables, z_init must have shape"
                        "[output_dim, num_inducing, input_dim]"
                    )
                z_init_o = z_init[o]
            else:
                z_init_o = np.random.randn(num_inducing, input_dim).astype(dtype=default_float())
            inducing_variables.append(InducingPoints(z_init_o))
        return SeparateIndependentInducingVariables(inducing_variables)

    else:
        # TODO: should we assert output_dim is None ?

        z_init = (
            z_init
            if z_init_is_given
            else np.random.randn(num_inducing, input_dim).astype(dtype=default_float())
        )
        shared_ip = InducingPoints(z_init)
        return SharedIndependentInducingVariables(shared_ip)


def construct_mean_function(
    X: np.ndarray, D_in: int, D_out: int
) -> gpflow.mean_functions.MeanFunction:
    """
    Return :class:`gpflow.mean_functions.Identity` when ``D_in`` and ``D_out`` are
    equal. Otherwise, use the principal components of the inputs matrix ``X`` to build a
    :class:`~gpflow.mean_functions.Linear` mean function.

    .. note::
        The returned mean function is set to be untrainable.
        To change this, use :meth:`gpflow.set_trainable`.

    :param X: A data array with the shape ``[N, D_in]`` used to determine the principal
        components to use to create a :class:`~gpflow.mean_functions.Linear` mean function
        when ``D_in != D_out``.
    :param D_in: The dimensionality of the input data (or features) ``X``.
        Typically, this corresponds to ``X.shape[-1]``.
    :param D_out: The dimensionality of the outputs (or targets) ``Y``.
        Typically, this corresponds to ``Y.shape[-1]`` or the number of latent GPs in the layer.
    """
    assert X.shape[-1] == D_in
    if D_in == D_out:
        mean_function = gpflow.mean_functions.Identity()
    else:
        if D_in > D_out:
            _, _, V = np.linalg.svd(X, full_matrices=False)
            W = V[:D_out, :].T
        else:
            W = np.concatenate([np.eye(D_in), np.zeros((D_in, D_out - D_in))], axis=1)

        assert W.shape == (D_in, D_out)
        mean_function = gpflow.mean_functions.Linear(W)
        gpflow.set_trainable(mean_function, False)

    return mean_function


def construct_gp_layer(
    num_data: int,
    num_inducing: int,
    input_dim: int,
    output_dim: int,
    kernel_class: Type[gpflow.kernels.Stationary] = gpflow.kernels.SquaredExponential,
    z_init: Optional[np.ndarray] = None,
    name: Optional[str] = None,
) -> GPLayer:
    """
    Builds a vanilla GP layer with a single kernel shared among all outputs,
        shared inducing point variables and zero mean function.

    :param num_data: total number of datapoints in the dataset, *N*.
        Typically corresponds to ``X.shape[0] == len(X)``.
    :param num_inducing: total number of inducing variables, *M*.
        This parameter can be freely chosen by the user. General advice
        is to pick it as high as possible, but smaller than *N*.
        The computational complexity of the layer is cubic in *M*.
    :param input_dim: dimensionality of the input data (or features) X.
        Typically, this corresponds to ``X.shape[-1]``.
    :param output_dim: The dimensionality of the outputs (or targets) ``Y``.
        Typically, this corresponds to ``Y.shape[-1]``.
    :param kernel_class: The kernel class used by the layer.
        This can be as simple as :class:`gpflow.kernels.SquaredExponential`, or more complex,
        for example, ``lambda **_: gpflow.kernels.Linear() + gpflow.kernels.Periodic()``.
        It will be passed a ``lengthscales`` keyword argument.
    :param z_init: The initial value for the inducing variable inputs.
    :param name: The name for the GP layer.
    """
    lengthscale = float(input_dim) ** 0.5
    base_kernel = kernel_class(lengthscales=np.full(input_dim, lengthscale))
    kernel = construct_basic_kernel(base_kernel, output_dim=output_dim, share_hyperparams=True)
    inducing_variable = construct_basic_inducing_variables(
        num_inducing,
        input_dim,
        output_dim=output_dim,
        share_variables=True,
        z_init=z_init,
    )
    gp_layer = GPLayer(
        kernel=kernel,
        inducing_variable=inducing_variable,
        num_data=num_data,
        mean_function=gpflow.mean_functions.Zero(),
        name=name,
    )
    return gp_layer


T = TypeVar("T")


def make_dataclass_from_class(dataclass: Type[T], instance: object, **updates: object) -> T:
    """
    Take a regular object ``instance`` with a superset of fields for a
    :class:`dataclasses.dataclass` (``@dataclass``-decorated class), and return an
    instance of the dataclass. The ``instance`` has all of the dataclass's fields
    but might also have more. ``key=value`` keyword arguments supersede the fields in ``instance``.
    """
    dataclass_keys = [f.name for f in fields(dataclass)]
    field_dict = {k: v for k, v in inspect.getmembers(instance) if k in dataclass_keys}
    field_dict.update(updates)
    return dataclass(**field_dict)  # type: ignore


def xavier_initialization_numpy(input_dim: int, output_dim: int) -> np.ndarray:
    r"""
    Generate initial weights for a neural network layer with the given input and output
    dimensionality using the Xavier Glorot normal initialiser. From:

    Glorot, Xavier, and Yoshua Bengio. "Understanding the difficulty of training deep
    feedforward neural networks." Proceedings of the thirteenth international
    conference on artificial intelligence and statistics. JMLR Workshop and Conference
    Proceedings, 2010.

    Draw samples from a normal distribution centred on :math:`0` with standard deviation
    :math:`\sqrt(2 / (\text{input_dim} + \text{output_dim}))`.
    """
    return np.random.randn(input_dim, output_dim) * (2.0 / (input_dim + output_dim)) ** 0.5



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides :func:`build_constant_input_dim_deep_gp` to build a Deep GP of
arbitrary depth where each hidden layer has the same input dimensionality as the data.
"""

from dataclasses import dataclass
from typing import cast

import numpy as np
import tensorflow as tf
from scipy.cluster.vq import kmeans2

import gpflow
from gpflow.kernels import SquaredExponential
from gpflow.likelihoods import Gaussian

from gpflux.helpers import (
    construct_basic_inducing_variables,
    construct_basic_kernel,
    construct_mean_function,
)
from gpflux.layers.gp_layer import GPLayer
from gpflux.layers.likelihood_layer import LikelihoodLayer
from gpflux.models import DeepGP


@dataclass
class Config:
    """
    The configuration used by :func:`build_constant_input_dim_deep_gp`.
    """

    num_inducing: int
    """
    The number of inducing variables, *M*. The Deep GP uses the same number
    of inducing variables in each layer.
    """

    inner_layer_qsqrt_factor: float
    """
    A multiplicative factor used to rescale the hidden layers'
    :attr:`~gpflux.layers.GPLayer.q_sqrt`. Typically this value is chosen to be small
    (e.g., 1e-5) to reduce noise at the start of training.
    """

    likelihood_noise_variance: float
    """
    The variance of the :class:`~gpflow.likelihoods.Gaussian` likelihood that is used
    by the Deep GP.
    """

    whiten: bool = True
    """
    Determines the parameterisation of the inducing variables.
    If `True`, :math:``p(u) = N(0, I)``, otherwise :math:``p(u) = N(0, K_{uu})``.
    .. seealso:: :attr:`gpflux.layers.GPLayer.whiten`
    """


def _construct_kernel(input_dim: int, is_last_layer: bool) -> SquaredExponential:
    """
    Return a :class:`gpflow.kernels.SquaredExponential` kernel with ARD lengthscales set to
    2 and a small kernel variance of 1e-6 if the kernel is part of a hidden layer;
    otherwise, the kernel variance is set to 1.0.

    :param input_dim: The input dimensionality of the layer.
    :param is_last_layer: Whether the kernel is part of the last layer in the Deep GP.
    """
    variance = 1e-6 if not is_last_layer else 1.0

    # TODO: Looking at this initializing to 2 (assuming N(0, 1) or U[0,1] normalized
    # data) seems a bit weird - that's really long lengthscales? And I remember seeing
    # something where the value scaled with the number of dimensions before
    lengthscales = [2.0] * input_dim
    return SquaredExponential(lengthscales=lengthscales, variance=variance)


def build_constant_input_dim_deep_gp(X: np.ndarray, num_layers: int, config: Config) -> DeepGP:
    r"""
    Build a Deep GP consisting of ``num_layers`` :class:`GPLayer`\ s.
    All the hidden layers have the same input dimension as the data, that is, ``X.shape[1]``.

    The architecture is largely based on :cite:t:`salimbeni2017doubly`, with
    the most notable difference being that we keep the hidden dimension equal
    to the input dimensionality of the data.

    .. note::
        This architecture might be slow for high-dimensional data.

    .. note::
        This architecture assumes a :class:`~gpflow.likelihoods.Gaussian` likelihood
        for regression tasks. Specify a different likelihood for performing
        other tasks such as classification.

    :param X: The training input data, used to retrieve the number of datapoints and
        the input dimension and to initialise the inducing point locations using k-means. A
        tensor of rank two with the dimensions ``[num_data, input_dim]``.
    :param num_layers: The number of layers in the Deep GP.
    :param config: The configuration for (hyper)parameters. See :class:`Config` for details.
    """
    if X.dtype != gpflow.default_float():
        raise ValueError(
            f"X needs to have dtype according to gpflow.default_float() = {gpflow.default_float()} "
            f"however got X with {X.dtype} dtype."
        )

    num_data, input_dim = X.shape
    X_running = X

    gp_layers = []
    centroids, _ = kmeans2(X, k=config.num_inducing, minit="points")

    for i_layer in range(num_layers):
        is_last_layer = i_layer == num_layers - 1
        D_in = input_dim
        D_out = 1 if is_last_layer else input_dim

        # Pass in kernels, specify output dim (shared hyperparams/variables)

        inducing_var = construct_basic_inducing_variables(
            num_inducing=config.num_inducing, input_dim=D_in, share_variables=True, z_init=centroids
        )

        kernel = construct_basic_kernel(
            kernels=_construct_kernel(D_in, is_last_layer),
            output_dim=D_out,
            share_hyperparams=True,
        )

        assert config.whiten is True, "non-whitened case not implemented yet"

        if is_last_layer:
            mean_function = gpflow.mean_functions.Zero()
            q_sqrt_scaling = 1.0
        else:
            mean_function = construct_mean_function(X_running, D_in, D_out)
            X_running = mean_function(X_running)
            if tf.is_tensor(X_running):
                X_running = cast(tf.Tensor, X_running).numpy()
            q_sqrt_scaling = config.inner_layer_qsqrt_factor

        layer = GPLayer(
            kernel,
            inducing_var,
            num_data,
            mean_function=mean_function,
            name=f"gp_{i_layer}",
        )
        layer.q_sqrt.assign(layer.q_sqrt * q_sqrt_scaling)
        gp_layers.append(layer)

    likelihood = Gaussian(config.likelihood_noise_variance)
    return DeepGP(gp_layers, LikelihoodLayer(likelihood))



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
An "encoder" for parametrizing latent variables. Does not work with mini-batching.
"""

from typing import Any, Optional, Tuple

import numpy as np
import tensorflow as tf

from gpflow import Parameter, default_float
from gpflow.base import TensorType
from gpflow.utilities.bijectors import positive

from gpflux.exceptions import EncoderInitializationError
from gpflux.layers import TrackableLayer


class DirectlyParameterizedNormalDiag(TrackableLayer):
    """
    This class implements direct parameterisation of the Normally-distributed
    posterior of the latent variables. A mean and standard deviation to parameterise a
    mean-field Normal distribution for each latent variable is created and learned
    during training. This type of encoder is not computationally very efficient for
    larger datasets, but can greatly simplify training, because no neural network is
    required to learn an amortized mapping.

    ..note::
        No amortization is used; each datapoint element has an associated mean and
        standard deviation. This is not compatible with mini-batching.

    See :cite:t:`dutordoir2018cde` for a more thorough explanation of latent variable models
    and encoders.
    """

    means: Parameter
    """
    Each row contains the value of the mean for a latent variable in the model.
    ``means`` is a tensor of rank two with the shape ``[N, W]`` because we have the same number
    of latent variables as datapoints, and each latent variable is ``W``-dimensional.
    Consequently, the mean for each latent variable is also ``W``-dimensional.
    """

    stds: Parameter
    """
    Each row contains the value of the diagonal covariances for a latent variable.
    ``stds`` is a tensor of rank two with the shape ``[N, W]`` because we have the same number
    of latent variables as datapoints, and each latent variable is ``W``-dimensional.
    Consequently, the diagonal elements of the square covariance matrix for each latent variable
    is also ``W``-dimensional.

    Initialised to ``1e-5 * np.ones((N, W))``.
    """

    def __init__(self, num_data: int, latent_dim: int, means: Optional[np.ndarray] = None):
        """
        Directly parameterise the posterior of the latent variables associated with
        each datapoint with a diagonal multivariate Normal distribution. Note that across
        latent variables we assume a mean-field approximation.

        See :cite:t:`dutordoir2018cde` for a more thorough explanation of
        latent variable models and encoders.

        :param num_data: The number of datapoints, ``N``.
        :param latent_dim: The dimensionality of the latent variable, ``W``.
        :param means: The initialisation of the mean of the latent variable posterior
            distribution. (see :attr:`means`). If `None` (the default setting), set to
            ``np.random.randn(N, W) * 0.01``; otherwise, ``means`` should be an array of
            rank two with the shape ``[N, W]``.
        """
        super().__init__()
        if means is None:
            # break the symmetry in the means:
            means = 0.01 * np.random.randn(num_data, latent_dim)
        else:
            if np.any(means.shape != (num_data, latent_dim)):
                raise EncoderInitializationError(
                    f"means must have shape [num_data, latent_dim] = [{num_data}, {latent_dim}]; "
                    f"got {means.shape} instead."
                )

        # initialise distribution with a small standard deviation, as this has
        # been observed to help fitting:
        stds = 1e-5 * np.ones_like(means)

        # TODO: Rename to `scale` and `loc` to match tfp.distributions
        self.means = Parameter(means, dtype=default_float(), name="w_means")
        self.stds = Parameter(stds, transform=positive(), dtype=default_float(), name="w_stds")

    def call(
        self, inputs: Optional[TensorType] = None, *args: Any, **kwargs: Any
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        Return the posterior's mean (see :attr:`means`) and standard deviation (see :attr:`stds`).
        """
        if inputs is not None:
            tf.debugging.assert_shapes(
                [(self.means, ["N", "W"]), (self.stds, ["N", "W"]), (inputs, ["N", "D"])]
            )
        return self.means, self.stds



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" This module implements a latent variable layer for deep GPs. """

import abc
from typing import Optional, Tuple

import tensorflow as tf
import tensorflow_probability as tfp

from gpflow import default_float
from gpflow.base import TensorType

from gpflux.layers.trackable_layer import TrackableLayer
from gpflux.types import ObservationType


class LayerWithObservations(TrackableLayer, metaclass=abc.ABCMeta):
    """
    By inheriting from this class, Layers indicate that their :meth:`call`
    method takes a second *observations* argument after the customary
    *layer_inputs* argument.

    This is used to distinguish which layers (unlike most standard Keras
    layers) require the original inputs and/or targets during training.
    For example, it is used by the amortized variational inference in the
    :class:`LatentVariableLayer`.
    """

    @abc.abstractmethod
    def call(
        self,
        layer_inputs: TensorType,
        observations: Optional[ObservationType] = None,
        training: Optional[bool] = None,
    ) -> tf.Tensor:
        """
        The :meth:`call` method of `LayerWithObservations` subclasses should
        accept a second argument, *observations*. In training mode, this will
        be the ``[inputs, targets]`` of the training points; otherwise, it is `None`.
        """


class LatentVariableLayer(LayerWithObservations):
    """
    A latent variable layer, with amortized mean-field variational inference.

    The latent variable is distribution-agnostic, but assumes a variational posterior
    that is fully factorised and is of the same distribution family as the prior.

    This class is used by models as described in :cite:p:`dutordoir2018cde, salimbeni2019iwvi`.
    """

    prior: tfp.distributions.Distribution
    """ The prior distribution for the latent variables. """

    encoder: tf.keras.layers.Layer
    """
    An encoder that maps from a concatenation of inputs and targets to the
    parameters of the approximate posterior distribution of the corresponding
    latent variables.
    """

    compositor: tf.keras.layers.Layer
    """
    A layer that takes as input the two-element ``[layer_inputs, latent_variable_samples]`` list
    and combines the elements into a single output tensor.
    """

    def __init__(
        self,
        prior: tfp.distributions.Distribution,
        encoder: tf.keras.layers.Layer,
        compositor: Optional[tf.keras.layers.Layer] = None,
        name: Optional[str] = None,
    ):
        """
        :param prior: A distribution that represents the :attr:`prior` over the latent variable.
        :param encoder: A layer which is passed the concatenated observation inputs
            and targets, and returns the appropriate parameters for the approximate
            posterior distribution; see :attr:`encoder`.
        :param compositor: A layer that combines layer inputs and latent variable
            samples into a single tensor; see :attr:`compositor`. If you do not specify a value for
            this parameter, the default is
            ``tf.keras.layers.Concatenate(axis=-1, dtype=default_float())``. Note that you should
            set ``dtype`` of the layer to GPflow's default dtype as in
            :meth:`~gpflow.default_float()`.
        :param name: The name of this layer (passed through to `tf.keras.layers.Layer`).
        """

        super().__init__(dtype=default_float(), name=name)
        self.prior = prior
        self.distribution_class = prior.__class__
        self.encoder = encoder
        self.compositor = (
            compositor
            if compositor is not None
            else tf.keras.layers.Concatenate(axis=-1, dtype=default_float())
        )

    def call(
        self,
        layer_inputs: TensorType,
        observations: Optional[ObservationType] = None,
        training: Optional[bool] = None,
        seed: Optional[int] = None,
    ) -> tf.Tensor:
        r"""
        Sample the latent variables and compose them with the layer input.

        When training, draw a sample of the latent variable from the posterior,
        whose distribution is parameterised by the encoder mapping from the data.
        Also add a KL divergence [posterior∥prior] to the losses.

        When not training, draw a sample of the latent variable from the prior.

        :param layer_inputs: The output of the previous layer.
        :param observations: The ``[inputs, targets]``, with the shapes ``[batch size, Din]``
            and ``[batch size, Dout]`` respectively. This parameter should be passed only when in
            training mode.
        :param training: The training mode indicator.
        :param seed: A random seed for the sampling operation.
        :returns: Samples of the latent variable composed with the layer inputs through the
            :attr:`compositor`
        """
        if training:
            if observations is None:
                raise ValueError("LatentVariableLayer requires observations when training")

            samples, loss_per_datapoint = self._inference_latent_samples_and_loss(
                layer_inputs, observations, seed=seed
            )
        else:
            samples = self._prediction_latent_samples(layer_inputs, seed=seed)
            loss_per_datapoint = tf.constant(0.0, dtype=default_float())

        self.add_loss(loss_per_datapoint)

        # Metric names should be unique; otherwise they get overwritten if you
        # have multiple with the same name
        name = f"{self.name}_local_kl" if self.name else "local_kl"
        self.add_metric(loss_per_datapoint, name=name, aggregation="mean")

        return self.compositor([layer_inputs, samples])

    def _inference_posteriors(
        self,
        observations: ObservationType,
        training: Optional[bool] = None,
    ) -> tfp.distributions.Distribution:
        """
        Return the posterior distributions parametrised by the :attr:`encoder`, which gets called
        with the concatenation of the inputs and targets in the *observations* argument.

        .. todo:: We might want to change encoders to have a
            `tfp.layers.DistributionLambda` final layer that directly returns the
            appropriately parameterised distributions object.

        :param observations: The ``[inputs, targets]``, with the shapes ``[batch size, Din]``
            and ``[batch size, Dout]`` respectively.
        :param training: The training mode indicator (passed through to the :attr:`encoder`'s call).
        :returns: The posterior distributions object.
        """
        inputs, targets = observations
        encoder_inputs = tf.concat(observations, axis=-1)

        distribution_params = self.encoder(encoder_inputs, training=training)
        posteriors = self.distribution_class(*distribution_params, allow_nan_stats=False)

        return posteriors

    def _inference_latent_samples_and_loss(
        self, layer_inputs: TensorType, observations: ObservationType, seed: Optional[int] = None
    ) -> Tuple[tf.Tensor, tf.Tensor]:
        r"""
        Sample latent variables during the *training* forward pass, hence requiring
        the observations. Also return the KL loss per datapoint.

        :param layer_inputs: The output of the previous layer _(unused)_.
        :param observations: The ``[inputs, targets]``, with the shapes ``[batch size, Din]``
            and ``[batch size, Dout]`` respectively.
        :param seed: A random seed for the sampling operation.
        :returns: The samples and the loss-per-datapoint.
        """
        posteriors = self._inference_posteriors(observations, training=True)
        samples = posteriors.sample(seed=seed)  # [N, Dw]
        # closed-form expectation E_q[log(q/p)] = KL[q∥p]:
        local_kls = self._local_kls(posteriors)
        loss_per_datapoint = tf.reduce_mean(local_kls, name="local_kls")
        return samples, loss_per_datapoint

    def _prediction_latent_samples(
        self, layer_inputs: TensorType, seed: Optional[int] = None
    ) -> tf.Tensor:
        """
        Sample latent variables during the *prediction* forward pass, only
        depending on the shape of this layer's inputs.

        :param layer_inputs: The output of the previous layer (for determining batch shape).
        :param seed: A random seed for the sampling operation.
        :returns: The samples.
        """
        batch_shape = tf.shape(layer_inputs)[:-1]
        samples = self.prior.sample(batch_shape, seed=seed)
        return samples

    def _local_kls(self, posteriors: tfp.distributions.Distribution) -> tf.Tensor:
        """
        Compute the KL divergences [posteriors∥prior].

        :param posteriors: A distribution that represents the approximate posteriors.
        :returns: The KL divergences from the prior for each of the posteriors.
        """
        return posteriors.kl_divergence(self.prior)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
TensorBoard event iterator
"""
from dataclasses import dataclass
from typing import Any, Iterator, List, Type, Union

import tensorflow as tf
from tensorflow.core.util import event_pb2
from tensorflow.python.framework import tensor_util


@dataclass
class Event:
    """Minimal container to hold TensorBoard event data"""

    tag: str  # summary name, e.g. "loss" or "lengthscales"
    step: int
    value: Any
    dtype: Type


def tensorboard_event_iterator(file_pattern: Union[str, List[str], tf.Tensor]) -> Iterator[Event]:
    """
    Iterator yielding preprocessed tensorboard Events.

    :param file_pattern: A string, a list of strings, or a `tf.Tensor` of string type
        (scalar or vector), representing the filename glob (i.e. shell wildcard)
        pattern(s) that will be matched.
    """

    def get_scalar_value(value: Any) -> Any:
        # Note(Vincent): I'm sorry this is messy...
        # Using `value.simple_value` returns 0.0 for
        # np.ndarray values, so we need to try `MakeNdarray`
        # first, which breaks for non-tensors.
        try:
            v = tensor_util.MakeNdarray(value.tensor).item()
        except Exception:
            try:
                v = value.simple_value
            except Exception:
                raise ValueError("Unable to read value from tensor")
        return v

    event_files = tf.data.Dataset.list_files(file_pattern)
    serialized_examples = tf.data.TFRecordDataset(event_files)
    for serialized_example in serialized_examples:
        event = event_pb2.Event.FromString(serialized_example.numpy())
        for value in event.summary.value:
            v = get_scalar_value(value)
            yield Event(tag=value.tag, step=event.step, value=v, dtype=type(v))



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" This module enables you to sample from (Deep) GPs using different approaches. """

import abc
from typing import Callable, Optional, Union

import tensorflow as tf

from gpflow.base import TensorType
from gpflow.conditionals import conditional
from gpflow.config import default_float, default_jitter
from gpflow.covariances import Kuf, Kuu
from gpflow.inducing_variables import InducingVariables
from gpflow.kernels import Kernel
from gpflow.utilities import Dispatcher

from gpflux.math import compute_A_inv_b
from gpflux.sampling.kernel_with_feature_decomposition import KernelWithFeatureDecomposition
from gpflux.sampling.utils import draw_conditional_sample

efficient_sample = Dispatcher("efficient_sample")
""" A function that returns a :class:`Sample` of a GP posterior. """


class Sample(abc.ABC):
    """
    This class represents a sample from a GP that you can evaluate by using the ``__call__``
    at new locations within the support of the GP.

    Importantly, the same function draw (sample) is evaluated when calling it multiple
    times. This property is called consistency. Achieving consistency for vanilla GPs is costly
    because it scales cubically with the number of evaluation points,
    but works with any kernel. It is implemented in
    :meth:`_efficient_sample_conditional_gaussian`.
    For :class:`KernelWithFeatureDecomposition`, the more efficient approach
    following :cite:t:`wilson2020efficiently` is implemented in
    :meth:`_efficient_sample_matheron_rule`.

    See the tutorial notebooks `Efficient sampling
    <../../../../notebooks/efficient_sampling.ipynb>`_ and `Weight Space
    Approximation with Random Fourier Features
    <../../../../notebooks/weight_space_approximation.ipynb>`_ for an
    in-depth overview.
    """

    @abc.abstractmethod
    def __call__(self, X: TensorType) -> tf.Tensor:
        r"""
        Return the evaluation of the GP sample :math:`f(X)` for :math:`f \sim GP(0, k)`.

        :param X: The inputs, a tensor with the shape ``[N, D]``, where ``D`` is the
            input dimensionality.
        :return: Function values, a tensor with the shape ``[N, P]``, where ``P`` is the
            output dimensionality.
        """
        raise NotImplementedError

    def __add__(self, other: Union["Sample", Callable[[TensorType], TensorType]]) -> "Sample":
        """
        Allow for the summation of two instances that implement the ``__call__`` method.
        """
        this = self.__call__

        class AddSample(Sample):
            def __call__(self, X: TensorType) -> tf.Tensor:
                return this(X) + other(X)

        return AddSample()


@efficient_sample.register(InducingVariables, Kernel, object)
def _efficient_sample_conditional_gaussian(
    inducing_variable: InducingVariables,
    kernel: Kernel,
    q_mu: tf.Tensor,
    *,
    q_sqrt: Optional[TensorType] = None,
    whiten: bool = False,
) -> Sample:
    """
    Most costly implementation for obtaining a consistent GP sample.
    However, this method can be used for any kernel.
    """

    class SampleConditional(Sample):
        # N_old is 0 at first, we then start keeping track of past evaluation points.
        X = None  # [N_old, D]
        P = tf.shape(q_mu)[-1]  # num latent GPs
        f = tf.zeros((0, P), dtype=default_float())  # [N_old, P]

        def __call__(self, X_new: TensorType) -> tf.Tensor:
            N_old = tf.shape(self.f)[0]
            N_new = tf.shape(X_new)[0]

            if self.X is None:
                self.X = X_new
            else:
                self.X = tf.concat([self.X, X_new], axis=0)

            mean, cov = conditional(
                self.X,
                inducing_variable,
                kernel,
                q_mu,
                q_sqrt=q_sqrt,
                white=whiten,
                full_cov=True,
            )  # mean: [N_old+N_new, P], cov: [P, N_old+N_new, N_old+N_new]
            mean = tf.linalg.matrix_transpose(mean)  # [P, N_old+N_new]
            f_old = tf.linalg.matrix_transpose(self.f)  # [P, N_old]
            f_new = draw_conditional_sample(mean, cov, f_old)  # [P, N_new]
            f_new = tf.linalg.matrix_transpose(f_new)  # [N_new, P]
            self.f = tf.concat([self.f, f_new], axis=0)  # [N_old + N_new, P]

            tf.debugging.assert_equal(tf.shape(self.f), [N_old + N_new, self.P])
            tf.debugging.assert_equal(tf.shape(f_new), [N_new, self.P])

            return f_new

    return SampleConditional()


@efficient_sample.register(InducingVariables, KernelWithFeatureDecomposition, object)
def _efficient_sample_matheron_rule(
    inducing_variable: InducingVariables,
    kernel: KernelWithFeatureDecomposition,
    q_mu: tf.Tensor,
    *,
    q_sqrt: Optional[TensorType] = None,
    whiten: bool = False,
) -> Sample:
    """
    Implements the efficient sampling rule from :cite:t:`wilson2020efficiently` using
    the Matheron rule. To use this sampling scheme, the GP has to have a
    ``kernel`` of the :class:`KernelWithFeatureDecomposition` type .

    :param kernel: A kernel of the :class:`KernelWithFeatureDecomposition` type, which
        holds the covariance function and the kernel's features and
        coefficients.
    :param q_mu: A tensor with the shape ``[M, P]``.
    :param q_sqrt: A tensor with the shape ``[P, M, M]``.
    :param whiten: Determines the parameterisation of the inducing variables.
    """
    L = tf.shape(kernel.feature_coefficients)[0]  # num eigenfunctions  # noqa: F841
    M, P = tf.shape(q_mu)[0], tf.shape(q_mu)[1]  # num inducing, num output heads

    prior_weights = tf.sqrt(kernel.feature_coefficients) * tf.random.normal(
        (L, P), dtype=default_float()
    )  # [L, P]

    u_sample_noise = tf.matmul(
        q_sqrt,
        tf.random.normal((P, M, 1), dtype=default_float()),  # [P, M, M]  # [P, M, 1]
    )  # [P, M, 1]
    Kmm = Kuu(inducing_variable, kernel, jitter=default_jitter())  # [M, M]
    tf.debugging.assert_equal(tf.shape(Kmm), [M, M])
    u_sample = q_mu + tf.linalg.matrix_transpose(u_sample_noise[..., 0])  # [M, P]

    if whiten:
        Luu = tf.linalg.cholesky(Kmm)  # [M, M]
        u_sample = tf.matmul(Luu, u_sample)  # [M, P]

    phi_Z = kernel.feature_functions(inducing_variable.Z)  # [M, L]
    weight_space_prior_Z = phi_Z @ prior_weights  # [M, P]
    diff = u_sample - weight_space_prior_Z  # [M, P]
    v = compute_A_inv_b(Kmm, diff)  # [M, P]
    tf.debugging.assert_equal(tf.shape(v), [M, P])

    class WilsonSample(Sample):
        def __call__(self, X: TensorType) -> tf.Tensor:
            """
            :param X: evaluation points [N, D]
            :return: function value of sample [N, P]
            """
            N = tf.shape(X)[0]
            phi_X = kernel.feature_functions(X)  # [N, L]
            weight_space_prior_X = phi_X @ prior_weights  # [N, P]
            Knm = tf.linalg.matrix_transpose(Kuf(inducing_variable, kernel, X))  # [N, M]
            function_space_update_X = Knm @ v  # [N, P]

            tf.debugging.assert_equal(tf.shape(weight_space_prior_X), [N, P])
            tf.debugging.assert_equal(tf.shape(function_space_update_X), [N, P])

            return weight_space_prior_X + function_space_update_X  # [N, P]

    return WilsonSample()



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import List, Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np

from gpflow.base import TensorType


def plot_layer(
    X: TensorType,
    layer_input: TensorType,
    mean: List[TensorType],
    cov: List[TensorType],
    sample: List[TensorType],
    idx: Optional[int],
    axes: Optional[Sequence[plt.Axes]] = None,
) -> None:
    """
    :param X: original inputs to the DGP, shape [N, 1]
    :param layer_input: inputs to this layer, shape [N, 1]
    :param mean: mean of this layer's output, shape [N, 1]
    :param cov: covariance of this layer's output, shape [1, N, N]
    :param sample: samples from this layer's output, shape [S, N, 1]
    :param idx: the index of this layer (for labels)
    :param axes: the sequence of 3 axes on which to plot
    """
    if axes is None:
        fig, (ax1, ax2, ax3) = plt.subplots(3, 1, figsize=(3, 10))
    else:
        assert len(axes) == 3
        ax1, ax2, ax3 = axes

    # Input
    title = "Input"
    if idx is not None:
        title = f"Layer {idx + 1}\n{title}"
    ax1.set_title(title)
    ax1.plot(X, layer_input)

    # covariance
    ax2.matshow(np.squeeze(cov, axis=0), aspect="auto")
    ax2.set_yticklabels([])
    ax2.set_xticklabels([])

    # samples
    ax3.set_title("Samples")
    ax3.plot(X, np.squeeze(sample, axis=-1).T)


def plot_layers(
    X: TensorType, means: List[TensorType], covs: List[TensorType], samples: List[TensorType]
) -> None:  # pragma: no cover
    L = len(means)
    fig, axes = plt.subplots(3, L, figsize=(L * 3.33, 10))
    for i in range(L):
        layer_input = X if i == 0 else samples[i - 1][0]
        plot_layer(X, layer_input, means[i], covs[i], samples[i], i, axes[:, i])



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides the `LikelihoodLoss` adapter to use GPflow's
:class:`~gpflow.likelihoods.Likelihood` implementations as a
`tf.keras.losses.Loss`.
"""
from typing import Union

import tensorflow as tf
import tensorflow_probability as tfp

import gpflow
from gpflow.base import TensorType

from gpflux.types import unwrap_dist


class LikelihoodLoss(tf.keras.losses.Loss):
    r"""
    This class is a `tf.keras.losses.Loss` implementation that wraps a GPflow
    :class:`~gpflow.likelihoods.Likelihood` instance.

    When the prediction (last-layer output) is a
    :class:`~tfp.distributions.Distribution` ``q(f)``, calling this loss
    returns the negative variational expectation :math:`-\mathbb{E}_{q(f)}[\log
    p(y|f)]`. When the prediction is a `tf.Tensor`, calling this loss returns
    the negative log-probability :math:`-\log p(y|f)`.

    When you use this loss function in training a Keras model, the value of
    this loss is not logged explicitly (in contrast, the layer-specific losses
    are logged, as is the overall model loss).
    To output this loss value explicitly, wrap this class in a
    `tf.keras.metrics.Metric` and add it to the model metrics.

    .. note::

        Use **either** this `LikelihoodLoss` (e.g. together with a
        `tf.keras.Sequential` model) **or**
        :class:`~gpflux.layers.LikelihoodLayer` (together with
        `gpflux.models.DeepGP`). Do **not** use both at once because this would
        add the loss twice.
    """

    def __init__(self, likelihood: gpflow.likelihoods.Likelihood):
        """
        :param likelihood: the GPflow likelihood object to use.

            .. note:: If you want to train any parameters of the likelihood
                (e.g. likelihood variance), you must include the likelihood as
                an attribute on a :class:`~gpflux.layers.TrackableLayer`
                instance that is part of your model. (This is not required when
                instead you use a :class:`gpflux.layers.LikelihoodLayer`
                together with :class:`gpflux.models.DeepGP`.)
        """
        super().__init__()

        self.likelihood = likelihood

    def call(
        self,
        y_true: TensorType,
        f_prediction: Union[TensorType, tfp.distributions.MultivariateNormalDiag],
    ) -> tf.Tensor:
        """
        Note that we deviate from the Keras Loss interface by calling the
        second argument *f_prediction* rather than *y_pred*.
        """
        no_X = None
        if isinstance(unwrap_dist(f_prediction), tfp.distributions.MultivariateNormalDiag):

            F_mu = f_prediction.loc
            F_var = f_prediction.scale.diag ** 2
            return -self.likelihood.variational_expectations(no_X, F_mu, F_var, y_true)
        else:  # Tensor
            f_samples = f_prediction
            return -self.likelihood.log_prob(no_X, f_samples, y_true)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" This module provides the base implementation for DeepGP models. """

import itertools
from typing import List, Optional, Tuple, Type, Union

import tensorflow as tf

import gpflow
from gpflow.base import Module, TensorType

import gpflux
from gpflux.layers import LayerWithObservations, LikelihoodLayer
from gpflux.sampling.sample import Sample


class DeepGP(Module):
    """
    This class combines a sequential function model ``f(x) = fₙ(⋯ (f₂(f₁(x))))``
    and a likelihood ``p(y|f)``.

    Layers might depend on both inputs x and targets y during training by
    inheriting from :class:`~gpflux.layers.LayerWithObservations`; those will
    be passed the argument ``observations=[inputs, targets]``.

    When data is used with methods in this class (e.g. :meth:`predict_f` method), it needs to
    be with ``dtype`` corresponding to GPflow's default dtype as in :meth:`~gpflow.default_float()`.

    .. note:: This class is **not** a `tf.keras.Model` subclass itself. To access
       Keras features, call either :meth:`as_training_model` or :meth:`as_prediction_model`
       (depending on the use-case) to create a `tf.keras.Model` instance. See the method docstrings
       for more details.
    """

    inputs: tf.keras.Input
    targets: tf.keras.Input

    f_layers: List[tf.keras.layers.Layer]
    """ A list of all layers in this DeepGP (just :attr:`likelihood_layer` is separate). """

    likelihood_layer: gpflux.layers.LikelihoodLayer
    """ The likelihood layer. """

    default_model_class: Type[tf.keras.Model]
    """
    The default for the *model_class* argument of :meth:`as_training_model` and
    :meth:`as_prediction_model`. This must have the same semantics as `tf.keras.Model`,
    that is, it must accept a list of inputs and an output. This could be
    `tf.keras.Model` itself or `gpflux.optimization.NatGradModel` (but not, for
    example, `tf.keras.Sequential`).
    """

    num_data: int
    """
    The number of points in the training dataset. This information is used to
    obtain correct scaling between the data-fit and the KL term in the evidence
    lower bound (:meth:`elbo`).
    """

    def __init__(
        self,
        f_layers: List[tf.keras.layers.Layer],
        likelihood: Union[
            gpflux.layers.LikelihoodLayer, gpflow.likelihoods.Likelihood
        ],  # fully-qualified for autoapi
        *,
        input_dim: Optional[int] = None,
        target_dim: Optional[int] = None,
        default_model_class: Type[tf.keras.Model] = tf.keras.Model,
        num_data: Optional[int] = None,
    ):
        """
        :param f_layers: The layers ``[f₁, f₂, …, fₙ]`` describing the latent
            function ``f(x) = fₙ(⋯ (f₂(f₁(x))))``.
        :param likelihood: The layer for the likelihood ``p(y|f)``. If this is a
            GPflow likelihood, it will be wrapped in a :class:`~gpflux.layers.LikelihoodLayer`.
            Alternatively, you can provide a :class:`~gpflux.layers.LikelihoodLayer` explicitly.
        :param input_dim: The input dimensionality.
        :param target_dim: The target dimensionality.
        :param default_model_class: The default for the *model_class* argument of
            :meth:`as_training_model` and :meth:`as_prediction_model`;
            see the :attr:`default_model_class` attribute.
        :param num_data: The number of points in the training dataset; see the
            :attr:`num_data` attribute.
            If you do not specify a value for this parameter explicitly, it is automatically
            detected from the :attr:`~gpflux.layers.GPLayer.num_data` attribute in the GP layers.
        """
        self.inputs = tf.keras.Input((input_dim,), dtype=gpflow.default_float(), name="inputs")
        self.targets = tf.keras.Input((target_dim,), dtype=gpflow.default_float(), name="targets")
        self.f_layers = f_layers
        if isinstance(likelihood, gpflow.likelihoods.Likelihood):
            self.likelihood_layer = LikelihoodLayer(likelihood)
        else:
            self.likelihood_layer = likelihood
        self.default_model_class = default_model_class
        self.num_data = self._validate_num_data(f_layers, num_data)

    @staticmethod
    def _validate_num_data(
        f_layers: List[tf.keras.layers.Layer], num_data: Optional[int] = None
    ) -> int:
        """
        Check that the :attr:`~gpflux.layers.gp_layer.GPLayer.num_data`
        attributes of all layers in *f_layers* are consistent with each other
        and with the (optional) *num_data* argument.

        :returns: The validated number of datapoints.
        """
        for i, layer in enumerate(f_layers):
            layer_num_data = getattr(layer, "num_data", None)
            if num_data is None:
                num_data = layer_num_data
            else:
                if layer_num_data is not None and num_data != layer_num_data:
                    raise ValueError(
                        f"f_layers[{i}].num_data is inconsistent with num_data={num_data}"
                    )
        if num_data is None:
            raise ValueError("Could not determine num_data; please provide explicitly")
        return num_data

    @staticmethod
    def _validate_dtype(x: TensorType) -> None:
        """
        Check that data ``x`` is of correct ``dtype``, corresponding to GPflow's default dtype as
        defined by :meth:`~gpflow.default_float()`.

        :raise ValueError: If ``x`` is of incorrect ``dtype``.
        """
        if x.dtype != gpflow.default_float():
            raise ValueError(
                f"x needs to have dtype {gpflow.default_float()} (according to "
                f"gpflow.default_float()), however got x with {x.dtype} dtype."
            )

    def _evaluate_deep_gp(
        self,
        inputs: TensorType,
        targets: Optional[TensorType],
        training: Optional[bool] = None,
    ) -> tf.Tensor:
        """
        Evaluate ``f(x) = fₙ(⋯ (f₂(f₁(x))))`` on the *inputs* argument.

        Layers that inherit from :class:`~gpflux.layers.LayerWithObservations`
        are passed the additional keyword argument ``observations=[inputs,
        targets]`` if *targets* contains a value, or ``observations=None`` when
        *targets* is `None`.
        """
        features = inputs

        # NOTE: we cannot rely on the `training` flag here, as the correct
        # symbolic graph needs to be constructed at "build" time (before either
        # fit() or predict() get called).
        if targets is not None:
            observations = [inputs, targets]
        else:
            # TODO would it be better to simply pass [inputs, None] in this case?
            observations = None

        for layer in self.f_layers:
            if isinstance(layer, LayerWithObservations):
                features = layer(features, observations=observations, training=training)
            else:
                features = layer(features, training=training)
        return features

    def _evaluate_likelihood(
        self,
        f_outputs: TensorType,
        targets: Optional[TensorType],
        training: Optional[bool] = None,
    ) -> tf.Tensor:
        """
        Call the `likelihood_layer` on *f_outputs*, which adds the
        corresponding layer loss when training.
        """
        return self.likelihood_layer(f_outputs, targets=targets, training=training)

    def call(
        self,
        inputs: TensorType,
        targets: Optional[TensorType] = None,
        training: Optional[bool] = None,
    ) -> tf.Tensor:
        self._validate_dtype(inputs)
        if targets is not None:
            self._validate_dtype(targets)
        f_outputs = self._evaluate_deep_gp(inputs, targets=targets, training=training)
        y_outputs = self._evaluate_likelihood(f_outputs, targets=targets, training=training)
        return y_outputs

    def predict_f(self, inputs: TensorType) -> Tuple[tf.Tensor, tf.Tensor]:
        """
        :returns: The mean and variance (not the scale!) of ``f``, for compatibility with GPflow
           models.
        :raise ValueError: If ``x`` is of incorrect ``dtype``.

        .. note:: This method does **not** support ``full_cov`` or ``full_output_cov``.
        """
        self._validate_dtype(inputs)
        f_distribution = self._evaluate_deep_gp(inputs, targets=None)
        return f_distribution.loc, f_distribution.scale.diag ** 2

    def elbo(self, data: Tuple[TensorType, TensorType]) -> tf.Tensor:
        """
        :returns: The ELBO (not the per-datapoint loss!), for compatibility with GPflow models.
        """
        X, Y = data
        _ = self.call(X, Y, training=True)
        all_losses = [
            loss
            for layer in itertools.chain(self.f_layers, [self.likelihood_layer])
            for loss in layer.losses
        ]
        return -tf.reduce_sum(all_losses) * self.num_data

    def _get_model_class(self, model_class: Optional[Type[tf.keras.Model]]) -> Type[tf.keras.Model]:
        if model_class is not None:
            return model_class
        else:
            return self.default_model_class

    def as_training_model(
        self, model_class: Optional[Type[tf.keras.Model]] = None
    ) -> tf.keras.Model:
        r"""
        Construct a `tf.keras.Model` instance that requires you to provide both ``inputs``
        and ``targets`` to its call. This information is required for
        training the model, because the ``targets`` need to be passed to the `likelihood_layer` (and
        to :class:`~gpflux.layers.LayerWithObservations` instances such as
        :class:`~gpflux.layers.LatentVariableLayer`\ s, if present).

        When compiling the returned model, do **not** provide any additional
        losses (this is handled by the :attr:`likelihood_layer`).

        Train with

        .. code-block:: python

            model.compile(optimizer)  # do NOT pass a loss here
            model.fit({"inputs": X, "targets": Y}, ...)

        See `Keras's Endpoint layer pattern
        <https://keras.io/examples/keras_recipes/endpoint_layer_pattern/>`_
        for more details.

        .. note:: Use `as_prediction_model` if you want only to predict, and do not want to pass in
           a dummy array for the targets.

        :param model_class: The model class to use; overrides `default_model_class`.
        """
        model_class = self._get_model_class(model_class)
        outputs = self.call(self.inputs, self.targets)
        return model_class([self.inputs, self.targets], outputs)

    def as_prediction_model(
        self, model_class: Optional[Type[tf.keras.Model]] = None
    ) -> tf.keras.Model:
        """
        Construct a `tf.keras.Model` instance that requires only ``inputs``,
        which means you do not have to provide dummy target values when
        predicting at test points.

        Predict with

        .. code-block:: python

            model.predict(Xtest, ...)

        .. note:: The returned model will not support training; for that, use `as_training_model`.

        :param model_class: The model class to use; overrides `default_model_class`.
        """
        model_class = self._get_model_class(model_class)
        outputs = self.call(self.inputs)
        return model_class(self.inputs, outputs)


def sample_dgp(model: DeepGP) -> Sample:  # TODO: should this be part of a [Vanilla]DeepGP class?
    function_draws = [layer.sample() for layer in model.f_layers]
    # TODO: error check that all layers implement .sample()?

    class ChainedSample(Sample):
        """This class chains samples from consecutive layers."""

        def __call__(self, X: TensorType) -> tf.Tensor:
            for f in function_draws:
                X = f(X)
            return X

    return ChainedSample()



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Math utilities
"""
import tensorflow as tf

from gpflow import default_jitter
from gpflow.base import TensorType


def _cholesky_with_jitter(cov: TensorType) -> tf.Tensor:
    """
    Compute the Cholesky of the covariance, adding jitter (determined by
    :func:`gpflow.default_jitter`) to the diagonal to improve stability.

    :param cov: full covariance with shape ``[..., N, D, D]``.
    """
    # cov [..., N, D, D]
    cov_shape = tf.shape(cov)
    batch_shape = cov_shape[:-2]
    D = cov_shape[-2]
    jittermat = default_jitter() * tf.eye(
        D, batch_shape=batch_shape, dtype=cov.dtype
    )  # [..., N, D, D]
    return tf.linalg.cholesky(cov + jittermat)  # [..., N, D, D]


def compute_A_inv_b(A: TensorType, b: TensorType) -> tf.Tensor:
    r"""
    Computes :math:`A^{-1} b` using the Cholesky of ``A`` instead of the explicit inverse,
    as this is often numerically more stable.

    :param A: A positive-definite matrix with shape ``[..., M, M]``.
        Can contain any leading dimensions (``...``) as long as they correspond
        to the leading dimensions in ``b``.
    :param b: Tensor with shape ``[..., M, D]``.
        Can contain any leading dimensions (``...``) as long as they correspond
        to the leading dimensions in ``A``.

    :returns: Tensor with shape ``[..., M, D]``.
        Leading dimensions originate from ``A`` and ``b``.
    """
    # A = L Lᵀ
    L = tf.linalg.cholesky(A)
    # A⁻¹ = L⁻ᵀ L⁻¹
    L_inv_b = tf.linalg.triangular_solve(L, b)
    A_inv_b = tf.linalg.triangular_solve(L, L_inv_b, adjoint=True)  # adjoint = transpose
    return A_inv_b



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Optimization-related modules, currently just contains the `NatGradModel`
and `NatGradWrapper` classes to integrate
`gpflow.optimizers.NaturalGradient` with Keras.
"""
from gpflux.optimization.keras_natgrad import NatGradModel, NatGradWrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from pathlib import Path

from setuptools import find_namespace_packages, setup

requirements = [
    "deprecated",
    "gpflow>=2.6.3",
    "numpy",
    "scipy",
    "tensorflow>=2.5.0,<2.11.0; platform_system!='Darwin' or platform_machine!='arm64'",
    # NOTE: Support of Apple Silicon MacOS platforms is in an experimental mode
    "tensorflow-macos>=2.5.0,<2.11.0; platform_system=='Darwin' and platform_machine=='arm64'",
    # NOTE: once we require tensorflow-probability>=0.12, we can remove our custom deepcopy handling
    "tensorflow-probability>=0.13.0,<0.19.0",
    "protobuf~=3.19.0"
]

with open("README.md", "r") as file:
    long_description = file.read()

with open(Path(__file__).parent / "gpflux" / "version.py", "r") as version_file:
    exec(version_file.read())

setup(
    name="gpflux",
    version=__version__,
    author="Secondmind Labs",
    author_email="gpflux@secondmind.ai",
    long_description=long_description,
    long_description_content_type="text/markdown",
    description="GPflux: Deep GP library",
    license="Apache License 2.0",
    keywords="Deep-Gaussian-processes",
    install_requires=requirements,
    packages=find_namespace_packages(include=["gpflux*"]),
    package_data={"gpflux": ["py.typed"]},
    project_urls={
        "Source on GitHub": "https://github.com/secondmind-labs/GPflux",
        "Documentation": "https://secondmind-labs.github.io/GPflux/",
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module contains utilities for sampling from multivariate Gaussian distributions.
"""
import tensorflow as tf

from gpflow.base import TensorType
from gpflow.conditionals.util import sample_mvn

from gpflux.math import _cholesky_with_jitter


def draw_conditional_sample(mean: TensorType, cov: TensorType, f_old: TensorType) -> tf.Tensor:
    r"""
    Draw a sample :math:`\tilde{f}_\text{new}` from the conditional
    multivariate Gaussian :math:`p(f_\text{new} | f_\text{old})`, where the
    parameters ``mean`` and ``cov`` are the mean and covariance matrix of the
    joint multivariate Gaussian over :math:`[f_\text{old}, f_\text{new}]`.

    :param mean: A tensor with the shape ``[..., D, N+M]`` with the mean of
        ``[f_old, f_new]``. For each ``[..., D]`` this is a stacked vector of the
        form:

        .. math::

            \begin{pmatrix}
                 \operatorname{mean}(f_\text{old}) \;[N] \\
                 \operatorname{mean}(f_\text{new}) \;[M]
            \end{pmatrix}

    :param cov: A tensor with the shape ``[..., D, N+M, N+M]`` with the covariance of
        ``[f_old, f_new]``. For each ``[..., D]``, there is a 2x2 block matrix of the form:

        .. math::

            \begin{pmatrix}
                 \operatorname{cov}(f_\text{old}, f_\text{old}) \;[N, N]
                   & \operatorname{cov}(f_\text{old}, f_\text{new}) \;[N, M] \\
                 \operatorname{cov}(f_\text{new}, f_\text{old}) \;[M, N]
                   & \operatorname{cov}(f_\text{new}, f_\text{new}) \;[M, M]
            \end{pmatrix}

    :param f_old: A tensor of observations with the shape ``[..., D, N]``,
        drawn from Normal distribution with mean
        :math:`\operatorname{mean}(f_\text{old}) \;[N]`, and covariance
        :math:`\operatorname{cov}(f_\text{old}, f_\text{old}) \;[N, N]`

    :return: A sample :math:`\tilde{f}_\text{new}` from the conditional normal
        :math:`p(f_\text{new} | f_\text{old})` with the shape ``[..., D, M]``.
    """
    N, D = tf.shape(f_old)[-1], tf.shape(f_old)[-2]  # noqa: F841
    M = tf.shape(mean)[-1] - N
    cov_old = cov[..., :N, :N]  # [..., D, N, N]
    cov_new = cov[..., -M:, -M:]  # [..., D, M, M]
    cov_cross = cov[..., :N, -M:]  # [..., D, N, M]
    L_old = _cholesky_with_jitter(cov_old)  # [..., D, N, N]
    A = tf.linalg.triangular_solve(L_old, cov_cross, lower=True)  # [..., D, N, M]
    var_new = cov_new - tf.matmul(A, A, transpose_a=True)  # [..., D, M, M]
    mean_new = mean[..., -M:]  # [..., D, M]
    mean_old = mean[..., :N]  # [..., D, N]
    mean_old_diff = (f_old - mean_old)[..., None]  # [..., D, N, 1]
    AM = tf.linalg.triangular_solve(L_old, mean_old_diff)  # [..., D, N, 1]
    mean_new = mean_new + (tf.matmul(A, AM, transpose_a=True)[..., 0])  # [..., D, M]
    return sample_mvn(mean_new, var_new, full_cov=True)



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" A kernel's features and coefficients using Random Fourier Features (RFF). """

from gpflux.layers.basis_functions.fourier_features.random.base import (
    RandomFourierFeatures,
    RandomFourierFeaturesCosine,
)
from gpflux.layers.basis_functions.fourier_features.random.orthogonal import (
    OrthogonalRandomFeatures,
)

__all__ = [
    "OrthogonalRandomFeatures",
    "RandomFourierFeatures",
    "RandomFourierFeaturesCosine",
]



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" This module provides custom exceptions used by GPflux. """


class GPLayerIncompatibilityException(Exception):
    r"""
    This exception is raised when :class:`~gpflux.layers.GPLayer` is
    misconfigured. This can be caused by multiple reasons, but common misconfigurations are:

    - Incompatible or wrong type of :class:`~gpflow.kernels.Kernel`,
      :class:`~gpflow.inducing_variables.InducingVariables` and/or
      :class:`~gpflow.mean_functions.MeanFunction`.
    - Incompatible number of latent GPs specified.
    """


class EncoderInitializationError(Exception):
    r"""
    This exception is raised by an encoder (e.g. :class:`DirectlyParameterizedNormalDiag`)
    when parameters are not initialised correctly.
    """



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" Runtime checks """
from typing import Optional, Tuple

from gpflow.inducing_variables import (
    FallbackSeparateIndependentInducingVariables,
    MultioutputInducingVariables,
)
from gpflow.kernels import MultioutputKernel
from gpflow.mean_functions import MeanFunction

from gpflux.exceptions import GPLayerIncompatibilityException


def verify_compatibility(
    kernel: MultioutputKernel,
    mean_function: MeanFunction,
    inducing_variable: MultioutputInducingVariables,
) -> Tuple[int, int]:
    """
    Checks that the arguments are all compatible with each other for use in a `GPLayer`.

    :param kernel: The multioutput kernel for the layer.
    :param inducing_variable: The inducing features for the layer.
    :param mean_function: The mean function applied to the inputs.
    :raises GPLayerIncompatibilityException: If an incompatibility is detected.
    :returns: number of inducing variables and number of latent GPs
    """
    # TODO: This function could be simplified by upstream enhancements to
    # GPflow: e.g. by adding an `output_dim` attribute to
    # MultioutputInducingVariable subclasses

    if not isinstance(inducing_variable, MultioutputInducingVariables):
        raise GPLayerIncompatibilityException(
            "`inducing_variable` must be a `gpflow.inducing_variables.MultioutputInducingVariables`"
        )
    if not isinstance(kernel, MultioutputKernel):
        raise GPLayerIncompatibilityException(
            "`kernel` must be a `gpflow.kernels.MultioutputKernel`"
        )
    if not isinstance(mean_function, MeanFunction):
        raise GPLayerIncompatibilityException(
            "`kernel` must be a `gpflow.mean_functions.MeanFunction`"
        )

    latent_inducing_points: Optional[int] = None
    if isinstance(inducing_variable, FallbackSeparateIndependentInducingVariables):
        latent_inducing_points = len(inducing_variable.inducing_variable_list)

    num_latent_gps = kernel.num_latent_gps

    if latent_inducing_points is not None:
        if latent_inducing_points != num_latent_gps:
            raise GPLayerIncompatibilityException(
                f"The number of latent GPs ({num_latent_gps}) does not match "
                f"the number of separate independent inducing_variables ({latent_inducing_points})"
            )

    num_inducing_points = inducing_variable.num_inducing  # currently the same for each dim
    return num_inducing_points, num_latent_gps



#
# Copyright (c) 2021 The GPflux Contributors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
This module provides a set of common utilities for kernel feature decompositions.
"""
import tensorflow as tf

import gpflow
from gpflow.base import TensorType


def _matern_number(kernel: gpflow.kernels.Kernel) -> int:
    if isinstance(kernel, gpflow.kernels.Matern52):
        p = 2
    elif isinstance(kernel, gpflow.kernels.Matern32):
        p = 1
    elif isinstance(kernel, gpflow.kernels.Matern12):
        p = 0
    else:
        raise NotImplementedError("Not a recognized Matern kernel")
    return p


def _bases_cosine(X: TensorType, W: TensorType, b: TensorType) -> TensorType:
    """
    Feature map for random Fourier features (RFF) as originally prescribed
    by Rahimi & Recht, 2007 :cite:p:`rahimi2007random`.
    See also :cite:p:`sutherland2015error` for additional details.
    """
    proj = tf.matmul(X, W, transpose_b=True) + b  # [N, M]
    return tf.cos(proj)  # [N, M]


def _bases_concat(X: TensorType, W: TensorType) -> TensorType:
    """
    Feature map for random Fourier features (RFF) as originally prescribed
    by Rahimi & Recht, 2007 :cite:p:`rahimi2007random`.
    See also :cite:p:`sutherland2015error` for additional details.
    """
    proj = tf.matmul(X, W, transpose_b=True)  # [N, M]
    return tf.concat([tf.sin(proj), tf.cos(proj)], axis=-1)  # [N, 2M]
